"""Naively, Reinforcement Learning algorithms can dish out an update of the MPC
parametrization at every time step, by leveraging only the current information. However,
as in Deep RL, it makes sense to enable the agent to store past experiences and
re-use them, or at least use them in a batched fashion, to improve the stability and
convergence learning process. :class:`ExperienceReplay` allows a learning agent to store
and sample, when performing an update, past experiences."""

from collections import deque
from collections.abc import Iterable, Iterator
from itertools import chain
from typing import Optional, TypeVar, Union

import numpy as np

from ..util.seeding import RngType

ExpType = TypeVar("ExpType")


class ExperienceReplay(deque[ExpType]):
    """Class for Reinforcement Learning agents' traning to save and sample experience
    transitions.

    The class inherits from :class:`deque`, adding a couple of simple functionalities to
    it for sampling transitions at random from past observed data (see :meth:`reset` and
    :meth:`sample`).

    Parameters
    ----------
    iterable : Iterable of ExpType, optional
        Initial items to be inserted in the container. By default, empty.
    maxlen : int, optional
        Maximum length/capacity of the memory. If ``None``, the deque has no maximum
        size, which is the default behaviour.
    sample_size : int or float, optional
        Size (as integer, or float percentage of ``maxlen``) of the experience replay
        items to draw when performing an update. By default, one item per sampling is
        drawn. If a float percentage, ``maxlen`` must be provided.
    include_latest : int or float, optional
        Size (as integer, or float percentage of ``sample_size``) dedicated to including
        the latest experience items. By default, ``0``, i.e., no last item is included.
    seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
        Seed for the :class:`numpy.random.Generator` used for sampling. By default,
        ``None``.

    Raises
    ------
    TypeError
        Raises if ``sample_size`` is a float (a percentage of the maximum length), but
        ``maxlen`` is ``None``, since it is impossible to compute the percentage of an
        unknown quantity.
    """

    def __init__(
        self,
        iterable: Iterable[ExpType] = (),
        maxlen: Optional[int] = None,
        sample_size: Union[int, float] = 1,
        include_latest: Union[int, float] = 0,
        seed: RngType = None,
    ) -> None:
        if isinstance(sample_size, float) and maxlen is None:
            raise TypeError(
                "Cannot compute the percentage of an unknown quantity (maxlen is None)."
            )
        super().__init__(iterable, maxlen=maxlen)
        self.sample_size = sample_size
        self.include_latest = include_latest
        self.reset(seed)

    def reset(self, seed: RngType = None) -> None:
        """Resets the seed of the :class:`numpy.random.Generator` used for sampling."""
        self.np_random = np.random.default_rng(seed)

    def sample(self) -> Iterator[ExpType]:
        """Samples the experience memory and yields the sampled items.

        Returns
        -------
        sample : iterator of ExpType
            An iterable sample is yielded.
        """
        L = len(self)
        n = self.sample_size
        last_n = self.include_latest
        if isinstance(n, float):
            n = int(self.maxlen * n)
        n = min(max(n, 0), L)
        if isinstance(last_n, float):
            last_n = int(n * last_n)
        last_n = min(max(last_n, 0), n)

        # get last n indices and the sampled indices from the remaining
        last = range(L - last_n, L)
        sampled = self.np_random.choice(range(L - last_n), n - last_n, False)
        yield from (self[i] for i in chain(sampled, last))



from abc import ABC
from typing import Any, Generic, TypeVar

from ...optim.gradient_based_optimizer import GradientBasedOptimizer, LrType
from .agent import SymType
from .learning_agent import LearningAgent

ExpType = TypeVar("ExpType")


class RlLearningAgent(
    LearningAgent[SymType, ExpType], ABC, Generic[SymType, ExpType, LrType]
):
    r"""Base abstract class for learning agents that employe gradient-based RL
    strategies to learn/improve the MPC policy. The only difference with the
    :class:`LearningAgent` is that this class accepts the RL task's discount factor and
    a gradient-based optimizer that dictates how the learnable parameters are updated.

    Parameters
    ----------
    discount_factor : float
        In RL, the factor that discounts future rewards in favor of immediate rewards.
        Usually denoted as :math:`\gamma`. It should satisfy :math:`\gamma \in (0, 1]`.
    optimizer : GradientBasedOptimizer
        A gradient-based optimizer (e.g., :class:`optim.GradientDescent`) to
        compute the updates of the learnable parameters, based on the current
        gradient-based RL algorithm.
    kwargs
        Additional arguments to be passed to :class:`LearningAgent`.
    """

    def __init__(
        self,
        discount_factor: float,
        optimizer: GradientBasedOptimizer[SymType, LrType],
        **kwargs: Any,
    ) -> None:
        self.discount_factor = discount_factor
        self.optimizer = optimizer
        super().__init__(**kwargs)
        self.optimizer.set_learnable_parameters(self._learnable_pars)

    def _establish_callback_hooks(self) -> None:
        super()._establish_callback_hooks()
        optim = self.optimizer
        optimizer_hook = optim.hook
        if optimizer_hook is not None:
            self._hook_callback(repr(optim), optimizer_hook, optim.step)



import logging
from collections.abc import Iterable, Iterator
from inspect import getmembers, isfunction
from itertools import chain
from operator import itemgetter
from typing import Callable, Optional, TypeVar

import numpy as np
import numpy.typing as npt
from gymnasium import Env

from ...agents.common.agent import Agent
from ...agents.common.learning_agent import LearningAgent
from ...core.callbacks import (
    AgentCallbackMixin,
    LearningAgentCallbackMixin,
    _failure_msg,
)
from ...util.iters import bool_cycle
from .wrapper import SymType, Wrapper

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")
_MANDATORY_CALLBACKS = {
    "on_mpc_failure",
    "on_validation_start",
    "on_validation_end",
    "on_update_failure",
    "on_training_start",
    "on_training_end",
}

_pred = lambda o: isfunction(o) and o.__name__.startswith("on_")
_AGENT_CALLBACKS = set(map(itemgetter(0), getmembers(AgentCallbackMixin, _pred)))
_LEARNING_AGENT_CALLBACKS = set.difference(
    set(map(itemgetter(0), getmembers(LearningAgentCallbackMixin, _pred))),
    _AGENT_CALLBACKS,
)
del _pred


def _generate_method_caller(m: Callable) -> Callable:
    """Returns a method that calls the given method `m`."""

    def method_caller(*args, **kwargs):
        return m(*args, **kwargs)

    return method_caller


class Log(Wrapper[SymType]):
    """A wrapper class for logging information about an agent.

    Parameters
    ----------
    agent : LearningAgent or inheriting
        Agent to wrap.
    log_name : str, optional
        Name of the logger. If not provided, the name of the agent is used.
    level : int, optional
        The logging level, by default :attr:`logging.INFO`.
    to_file : bool, optional
        Whether to write the log also to a file in the current directory. By
        default, ``False``.
    mode : str, optional
        The mode for opening the logging faile, in case ``to_file=True``. By default, it
        appends to the file, if already present.
    precision : int, optional
        Precision for printing floats, by default ``3``.
    log_frequencies : dict of (str, int), optional
        A dict containing, for each logging call hook, its corresponding frequency. The
        calls for which a frequency can be set are:

        - ``"on_episode_start"``
        - ``"on_episode_end"``
        - ``"on_env_step"``
        - ``"on_timestep_end"``
        - ``"on_update"``.

        If this dictionary does not contain an entry for a specific call, the call is
        assumed to be never logged.
    exclude_mandatory : iterable of str, optional
        An iterable of strings that contains the default mandatory callbacks to be
        excluded. These mandatory callbacks that can be excluded are:

        - ``"on_mpc_failure"``
        - ``"on_validation_start"``
        - ``"on_validation_end"``
        - ``"on_update_failure"``
        - ``"on_training_start"``
        - ``"on_training_end"``.
    """

    def __init__(
        self,
        agent: Agent[SymType],
        log_name: Optional[str] = None,
        level: int = logging.INFO,
        to_file: bool = False,
        mode: str = "a",
        precision: int = 3,
        log_frequencies: Optional[dict[str, int]] = None,
        exclude_mandatory: Optional[Iterable[str]] = None,
    ) -> None:
        name = log_name if log_name is not None else agent.name
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        formatter = logging.Formatter(
            fmt="%(name)s@%(asctime)s> %(message)s", datefmt="%Y-%m-%d,%H:%M:%S"
        )
        ch = logging.StreamHandler()
        ch.setLevel(level)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        if to_file:
            fh = logging.FileHandler(f"{name}.txt", mode=mode)
            fh.setLevel(level)
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)
        self.precision = precision

        # store excluded-mandatory-callbacks and callbacks-with-frequency
        self.exclude_mandatory: set[str] = (
            set() if exclude_mandatory is None else set(exclude_mandatory)
        )
        self.log_frequencies: dict[str, Iterator[bool]] = {}
        if log_frequencies is not None:
            for name, freq in log_frequencies.items():
                if name not in _MANDATORY_CALLBACKS:
                    self.log_frequencies[name] = bool_cycle(freq)

        # if the agent is non-learning, make sure that both mandatory and callbacks with
        # frequencies do not lead to callbacks reserved to only learning agents.
        if not isinstance(agent.unwrapped, LearningAgent):
            for cb in _LEARNING_AGENT_CALLBACKS:
                self.exclude_mandatory.add(cb)
                self.log_frequencies.pop(cb, None)
        super().__init__(agent)

    def _establish_callback_hooks(self) -> None:
        super()._establish_callback_hooks()
        # hook only the callbacks for which a frequency was given + the mandatory ones
        repr_self = repr(self)
        optional_cbs = self.log_frequencies.keys()
        mandatory_cbs = _MANDATORY_CALLBACKS.difference(self.exclude_mandatory)
        for name in chain(optional_cbs, mandatory_cbs):
            method = getattr(self, f"_{name}")
            self._hook_callback(repr_self, name, _generate_method_caller(method))

    # NOTE: callbacks for Agent

    def _on_mpc_failure(
        self, episode: int, timestep: Optional[int], status: str, raises: bool
    ) -> None:
        m = self.logger.error if raises else self.logger.warning
        m(_failure_msg("mpc", self.agent.name, episode, timestep, status))

    def _on_validation_start(self, env: Env[ObsType, ActType]) -> None:
        self.logger.debug("validation of %s started.", env)

    def _on_validation_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        S = np.array2string(returns, precision=self.precision)
        self.logger.info("validation of %s concluded with returns=%s.", env, S)

    def _on_episode_start(
        self, env: Env[ObsType, ActType], episode: int, state: ObsType
    ) -> None:
        if next(self.log_frequencies["on_episode_start"]):
            S = np.array2string(state, precision=self.precision)
            self.logger.debug("episode %d started with state=%s.", episode, S)

    def _on_episode_end(
        self, env: Env[ObsType, ActType], episode: int, rewards: float
    ) -> None:
        if next(self.log_frequencies["on_episode_end"]):
            self.logger.info(
                "episode %d ended with rewards=%.*f.", episode, self.precision, rewards
            )

    def _on_env_step(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        if next(self.log_frequencies["on_env_step"]):
            self.logger.debug(
                "env stepped in episode %d at time %d.", episode, timestep
            )

    def _on_timestep_end(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        if next(self.log_frequencies["on_timestep_end"]):
            self.logger.debug("episode %d stepped at time %d.", episode, timestep)

    # NOTE: callbacks for LearningAgent

    def _on_update_failure(
        self, episode: int, timestep: Optional[int], errormsg: str, raises: bool
    ) -> None:
        (self.logger.error if raises else self.logger.warning)(
            "_failure_msg('update', %s, %d, %s, %s)",
            self.agent.name,
            episode,
            timestep,
            errormsg,
        )

    def _on_training_start(self, env: Env[ObsType, ActType]) -> None:
        self.logger.debug("training of %s started.", env)

    def _on_training_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        S = np.array2string(returns, precision=self.precision)
        self.logger.info("training of %s concluded with returns=%s.", env, S)

    def _on_update(self) -> None:
        if next(self.log_frequencies["on_update"]):
            S = self.agent.learnable_parameters.stringify()
            self.logger.info("updated parameters: %s.", S)



"""As it will be clear from the inheritance diagram in :ref:`module_reference_agents`,
all agents are derived from mixin classes that define callbacks and manage hooks
attached to these callbacks. These system allows not only the user to customize the
behaviour of a derived agent every time a callback is triggered, but also to easily
implement and manage all those events and quantities that need to be scheduled during
training and evaluation. Some examples of such events are the decay of the learning rate
or the exploration chances, or when and with which frequency to invoke an update of the
MPC parametrization. Here we list the classes that enable this system, but for an
introduction to the callbacks and how to use them, see :ref:`user_guide_callbacks`."""

from typing import Any, Callable, Literal, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from gymnasium import Env

from .errors import raise_or_warn_on_mpc_failure, raise_or_warn_on_update_failure

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


def _failure_msg(
    category: Literal["mpc", "update"],
    name: str,
    episode: int,
    timestep: Optional[int],
    status: str,
) -> str:
    """Internal utility for composing message for mpc/update failure."""
    C = category.title()
    if timestep is None:
        return f"{C} failure of {name} at episode {episode}, status: {status}."
    else:
        return (
            f"{C} failure of {name} at episode {episode}, time {timestep}, "
            f"status: {status}."
        )


class CallbackMixin:
    """A class with the particular purpose of creating, storing and deleting hooks
    attached to callbacks.

    Notes
    -----
    A particular note must be included about the `__setstate__` method. When this method
    is used (e.g., via :func:`copy.deepcopy`), the hooks are not copied from the old
    copy. The reason is that the old copy/state's hooks are likely to be pointing to
    methods belonging to old objects' instances. Of course, this is an issue, because if
    the old hooks are used, the new object (created from the state) would reference
    callbacks belonging to the old object. For this reasons, hooks are not copied;
    instead, the method :meth:`_establish_callback_hooks` is automatically called to
    re-establish these, but with respect to the new object(s).
    """

    def __init__(self) -> None:
        self._hooks: dict[str, dict[str, Callable[..., None]]] = {}

    def __setstate__(
        self,
        state: Union[
            None, dict[str, Any], tuple[Optional[dict[str, Any]], dict[str, Any]]
        ],
    ) -> None:
        if isinstance(state, tuple) and len(state) == 2:
            state, slotstate = state
        else:
            slotstate = None
        if state is not None:
            # remove hooks (otherwise, new copies will still be calling the old object)
            state["_hooks"] = {}
            self.__dict__.update(state)
        if slotstate is not None:
            for key, value in slotstate.items():
                setattr(self, key, value)
        # re-establish hooks
        self._establish_callback_hooks()

    def _run_hooks(self, method_name: str, *args: Any) -> None:
        """Runs the internal hooks attached to the given method."""
        if hooks := self._hooks.get(method_name):
            for hook in hooks.values():
                hook(*args)

    def _establish_callback_hooks(self) -> None:
        """This method must be used to perform the connections between callbacks and any
        invokable method (hook). If the object has no hooks, then this method does
        nothing."""

    def _hook_callback(
        self, attachername: str, callbackname: str, func: Callable[..., None]
    ) -> None:
        """Hooks a function to be called each time a callback is invoked.

        Parameters
        ----------
        attachername : str
            The name of the object requesting the hook. Has only info purposes.
        callbackname : str
            Name of the callback to hook to, i.e., the target of the hooking.
        func : Callable
            function to be called when the callback is invoked. Must accept the same
            input arguments as the callback it is hooked to. The return value is
            discarded.

        Raises
        ------
        ValueError
            If an hook with name ``attachername`` is already attached to this callback.
        """
        hook_dict = self._hooks.setdefault(callbackname, {})
        if attachername in hook_dict:
            raise ValueError(
                f"Hook '{attachername}' already attached to callback '{callbackname}'."
            )
        hook_dict[attachername] = func


class AgentCallbackMixin(CallbackMixin):
    """Class with callbacks for agents.

    In particular, this class defines the following callbacks:

    - :meth:`on_mpc_failure`, invoked when an MPC solver fails
    - :meth:`on_validation_start`, invoked when validation starts (see
      :meth:`mpcrl.Agent.evaluate`)
    - :meth:`on_validation_end`, invoked when validation ends
    - :meth:`on_episode_start`, invoked when a training or validation episode starts
    - :meth:`on_episode_end`, invoked when a training or validation episode ends
    - :meth:`on_env_step`, invoked when a training or validation episode steps, i.e.,
      after :func:`gymnasium.Env.step`
    - :meth:`on_timestep_end`, invoked when the current simulation's time step reaches
      an end, i.e., after having stepped the environment and done all the internal
      computations according to the algorithm.
    """

    def on_mpc_failure(
        self, episode: int, timestep: Optional[int], status: str, raises: bool
    ) -> None:
        """Callback in case of failure of the MPC solver.

        Parameters
        ----------
        episode : int
            Number of the episode when the failure happened.
        timestep : int or None
            Timestep of the current episode when the failure happened. Can be ``None``,
            in case the error occurs inter-episodically or no notion of time step is
            available.
        status : str
            Status of the solver that failed.
        raises : bool
            Whether the failure should be raised as exception (``True``) or as a warning
            (``False``).
        """
        name: str = getattr(self, "name", "agent")
        raise_or_warn_on_mpc_failure(
            _failure_msg("mpc", name, episode, timestep, status),
            raises,
        )
        self._run_hooks("on_mpc_failure", episode, timestep, status, raises)

    def on_validation_start(self, env: Env[ObsType, ActType]) -> None:
        """Callback called at the beginning of the validation process (see
        :meth:`mpcrl.Agent.evaluate`)

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being validated on.
        """
        self._run_hooks("on_validation_start", env)

    def on_validation_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        """Callback called at the end of the validation process (see
        :meth:`mpcrl.Agent.evaluate`).

        Parameters
        ----------
        env : gym env
            A gym environment where the agent has been validated on.
        returns : array of double
            Each episode's cumulative rewards.
        """
        self._run_hooks("on_validation_end", env, returns)

    def on_episode_start(
        self, env: Env[ObsType, ActType], episode: int, state: ObsType
    ) -> None:
        """Callback called at the beginning of each episode in the training or
        validation process (see :meth:`mpcrl.Agent.evaluate`,
        :meth:`mpcrl.LearningAgent.train` and
        :meth:`mpcrl.LearningAgent.train_offpolicy`).

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        state : ObsType
            Starting state for this episode.
        """
        self._run_hooks("on_episode_start", env, episode, state)

    def on_episode_end(
        self, env: Env[ObsType, ActType], episode: int, rewards: float
    ) -> None:
        """Callback called at the end of each episode in the training or evaluation
        process (see :meth:`mpcrl.Agent.evaluate`, :meth:`mpcrl.LearningAgent.train` and
        :meth:`mpcrl.LearningAgent.train_offpolicy`).

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        rewards : float
            Cumulative rewards for this episode.
        """
        self._run_hooks("on_episode_end", env, episode, rewards)

    def on_env_step(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        """Callback called after each call to :func:`gymnasium.Env.step`.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        timestep : int
            Time instant of the current training episode.
        """
        self._run_hooks("on_env_step", env, episode, timestep)

    def on_timestep_end(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        """Callback called at the end of each time iteration. It is called with the same
        frequency as :meth:`on_env_step`, but with different timing.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        timestep : int
            Time instant of the current training episode.
        """
        self._run_hooks("on_timestep_end", env, episode, timestep)


class LearningAgentCallbackMixin(AgentCallbackMixin):
    """Class with callbacks for learning agents.

    In particular, this class defines, on top of the callbacks from
    :class:`AgentCallbackMixin`, the additional following callbacks:

    - :meth:`on_update_failure`, invoked when an update of the parametrization fails
    - :meth:`on_training_start`, invoked when training starts (see
      :meth:`mpcrl.LearningAgent.train` and :meth:`mpcrl.LearningAgent.train_offpolicy`)
    - :meth:`on_training_end`, invoked when training ends
    - :meth:`on_update`, invoked after each update of the parametrization.
    """

    def on_update_failure(
        self, episode: int, timestep: Optional[int], errormsg: str, raises: bool
    ) -> None:
        """Callback in case of update failure.

        Parameters
        ----------
        episode : int
            Number of the episode when the failure happened.
        timestep : int or None
            Timestep of the current episode when the failure happened. Can be ``None``
            in case the update occurs inter-episodically or no notion of time step is
            available.
        errormsg : str
            Error message of the update failure.
        raises : bool
            Whether the failure should be raised as exception (``True``) or as a warning
            (``False``).
        """
        name: str = getattr(self, "name", "agent")
        raise_or_warn_on_update_failure(
            _failure_msg("update", name, episode, timestep, errormsg),
            raises,
        )
        self._run_hooks("on_update_failure", episode, timestep, errormsg, raises)

    def on_training_start(self, env: Env[ObsType, ActType]) -> None:
        """Callback called at the beginning of the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        """
        self._run_hooks("on_training_start", env)

    def on_training_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        """Callback called at the end of the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent has been trained on.
        returns : array of double
            Each episode's cumulative rewards.
        """
        self._run_hooks("on_training_end", env, returns)

    def on_update(self) -> None:
        """Callback called after each :func:`mpcrl.LearningAgent.update`.

        This callback is especially useful for, e.g., decaying exploration probabilities
        or learning rates."""
        self._run_hooks("on_update")



"""The update strategy is likely to be one of the most important aspects that a designer
has to consider when training a Reinforcement Learning agent. When instantiating an
agent, via :class:`UpdateStrategy`, the user can specify when and with which frequency
to update the agent's MPC parametrization (e.g., at the end of every training episode,
or every ``N`` time steps), as well as the number of updates to skip at the beginning
(in case we need to wait for experience buffers to properly fill first)."""

from collections.abc import Iterator
from itertools import chain, repeat
from typing import Literal

from ..util.iters import bool_cycle


class UpdateStrategy:
    """A class holding information on the update strategy to be used by the learning
    algorithm.

    Parameters
    ----------
    frequency : int
        Frequency at which, each time the hook is called, an update should be
        carried out.
    skip_first : int, optional
        Skips the first ``skip_first`` updates. By default ``0``, so no update is
        skipped. This is useful when, e.g., the agent has to wait for the experience
        buffer to be filled before starting to update.
    hook : {"on_episode_end", "on_timestep_end"}, optional
        Specifies to which callback to hook, i.e., when to check if an update is due
        according to the given frequency. The options are:

        - ``"on_episode_end"`` checks if an update is due  after each episode ends
        - ``"on_timestep_end"`` checks for an update after each simulation's time step.

        By default, ``"on_timestep_end"`` is selected.
    """

    def __init__(
        self,
        frequency: int,
        hook: Literal["on_episode_end", "on_timestep_end"] = "on_timestep_end",
        skip_first: int = 0,
    ) -> None:
        self.frequency = frequency
        self.hook = hook
        self._update_cycle = chain(
            repeat(False, skip_first * frequency), bool_cycle(frequency)
        )

    def can_update(self) -> bool:
        """Returns whether an update must be carried out now, at the current instant,
        according to the specified strategy.

        Notes
        -----
        This methods steps the internal iterators to check whether an update is due with
        :func:`next`. This means that calling this method has a side effect on the state
        of these iterators, and calling immediately again can result in a different
        outcome.

        Returns
        -------
        bool
            ``True`` if the agent should update according to this strategy; otherwise,
            ``False``.
        """
        return next(self._update_cycle)

    def __iter__(self) -> Iterator[bool]:
        return self._update_cycle

    def __next__(self) -> bool:
        return next(self._update_cycle)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(frequency={self.frequency},hook={self.hook})"

    def __str__(self) -> str:
        return self.__repr__()



from typing import Any, Generic

import numpy as np

from ..core.parameters import LearnableParametersDict, SymType


class BaseOptimizer(Generic[SymType]):
    """Base class for optimization algorithms.

    This class contains useful methods for, e.g., initializing the optimizer, retrieving
    bounds on the learnable parameters, etc.

    Parameters
    ----------
    max_percentage_update : float, optional
        A positive float that specifies the maximum percentage change the learnable
        parameters can experience in each update. For example,
        ``max_percentage_update=0.5`` means that the parameters can be updated by up
        to 50% of their current value. By default, it is set to ``+inf``.
    """

    def __init__(self, max_percentage_update: float = float("+inf")) -> None:
        self.max_percentage_update = max_percentage_update
        self.learnable_parameters: LearnableParametersDict[SymType] = None

    def set_learnable_parameters(self, pars: LearnableParametersDict[SymType]) -> None:
        """Makes the optimization class aware of the dictionary of the learnable
        parameters whose values are to be updated.

        Parameters
        ----------
        pars : :class`mpcrl.LearnableParametersDict`
            The dictionary of the learnable parameters.
        """
        self.learnable_parameters = pars
        self._update_solver = self._init_update_solver()

    def _get_update_bounds(
        self, theta: np.ndarray, eps: float = 0.1
    ) -> tuple[np.ndarray, np.ndarray]:
        """Internal utility to retrieve the current bounds on the learnable parameters.
        Only useful if the update problem is not unconstrained, i.e., there are either
        some lower- or upper-bounds, or a maximum percentage update was given."""
        lb = self.learnable_parameters.lb - theta
        ub = self.learnable_parameters.ub - theta
        perc = self.max_percentage_update
        if perc != float("+inf"):
            max_update_delta = np.maximum(np.abs(perc * theta), eps)
            lb = np.maximum(lb, -max_update_delta)
            ub = np.minimum(ub, +max_update_delta)
        return lb, ub

    def _init_update_solver(self) -> Any:
        """Internal utility to initialize whatever solver is necessary to perform an
        update according to this learning strategy."""

    def __str__(self) -> str:
        return self.__class__.__name__

    def __repr__(self) -> str:
        cn = self.__class__.__name__
        mp = self.max_percentage_update
        return f"{cn}(max%={mp})"



from typing import Any, Literal, Optional, Union

import numpy as np
import numpy.typing as npt
from scipy.linalg import cho_solve

from ..core.schedulers import Scheduler
from ..util.math import cholesky_added_multiple_identity
from .gradient_based_optimizer import GradientBasedOptimizer, LrType, SymType


class NetwonMethod(GradientBasedOptimizer[SymType, LrType]):
    r"""Second-order gradient-based Netwon's method.

    In constrast to the first-order methods, the Netwon's method uses also the Hessian
    of the loss function to compute the update. The unconstrained update is given by

    .. math:: \theta \gets \theta - \alpha H^{-1} g.

    However, we do not directly use the provided Hessian, but rather its Cholesky
    decomposition after having ensured it is positive semi-definite via
    :meth:`cholesky_added_multiple_identity`. As usual, weight decay can be added, but
    for sake of simplicity it is not included in the formula above. In case there are
    constraints on the learnable parameters, the update is solved as a Quadratic
    Programming (QP) problem, which is slower than the unconstrained counterpart. This
    QP takes the form

    .. math::

        \begin{aligned}
            \min_{\Delta\theta} & \quad \frac{1}{2} \Delta\theta^\top H \Delta\theta + \alpha g^\top \Delta\theta \\
            \text{s.t.} & \quad \theta_{\text{lower}} \leq \theta + \Delta\theta \leq \theta_{\text{upper}}
        \end{aligned}

    if ``cho_before_update=False``; otherwise, the objective is
    :math:`\frac{1}{2} \lVert \Delta\theta \rVert_2^2 + \alpha (H^{-1} g)^\top \Delta\theta`.

    Parameters
    ----------
    learning_rate : float or array or :class:`mpcrl.core.schedulers.Scheduler`
        The learning rate of the optimizer. It can be:

        - a float, in case the learning rate must stay constant and is the same for all
          learnable parameters

        - an array, in case the learning rate must stay constant but is different for
          each parameter (should have the same size as the number of learnable
          parameters)

        - a :class:`mpcrl.core.schedulers.Scheduler`, in case the learning rate can vary
          during the learning process (usually, it is set to decay). See the ``hook``
          argument for more details on when this scheduler is stepped.
    weight_decay : float, optional
        A positive float that specifies the decay of the learnable parameters in the
        form of an L2 regularization term. By default, it is set to ``0.0``, so no
        decay/regularization takes place.
    cho_before_update : bool, optional
        Whether to perform a Cholesky's factorization of the hessian in preparation
        of each update. If ``False``, the quadratic form in the QP objective hosts the
        Hessian matrix; else if ``True``, the linear system :math:`H^{-1} g` is first
        solved via Cholesky's factorization, and the QP update's Hessian is downgraded
        to an identity matrix. Only relevant if the update is constrained. By
        default, ``False``.
    cho_maxiter : int, optional
        Maximum number of iterations in the Cholesky's factorization with additive
        multiples of the identity to ensure positive definiteness of the hessian. By
        default, ``1000``.
    cho_solve_kwargs : kwargs for :func:`scipy.linalg.cho_solve`, optional
        The optional kwargs to be passed to :func:`scipy.linalg.cho_solve` to solve
        linear systems with the Hessian's Cholesky decomposition. If ``None``, it is set
        by default to ``cho_solve_kwargs = {"check_finite": False }``. Only relevant if
        no weight decay is given.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies when to step the optimizer's learning rate's scheduler to decay
        its value. This allows to vary the rate over the learning iterations. The
        options are:

        - ``"on_update"`` steps the learning rate after each agent's update

        - ``"on_episode_end"`` steps the learning rate after each episode's end

        - ``"on_timestep_end"`` steps the learning rate after each env's timestep.

        By default, ``"on_update"`` is selected.
    max_percentage_update : float, optional
        A positive float that specifies the maximum percentage change the learnable
        parameters can experience in each update. For example,
        ``max_percentage_update=0.5`` means that the parameters can be updated by up
        to 50% of their current value. By default, it is set to ``+inf``.
        If specified, the update becomes constrained and has to be solved as a QP, which
        is inevitably slower than its unconstrained counterpart (a linear system).
    bound_consistency : bool, optional
        A boolean that, if ``True``, forces the learnable parameters to lie in their
        bounds when updated. This is done via :func:`numpy.clip`. Only beneficial if
        numerical issues arise during updates, e.g., due to the QP solver not being able
        to guarantee bounds.
    """

    _order = 2
    _hessian_sparsity = "dense"

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        weight_decay: float = 0.0,
        cho_before_update: bool = False,
        cho_maxiter: int = 1000,
        cho_solve_kwargs: Optional[dict[str, Any]] = None,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
        bound_consistency: bool = False,
    ) -> None:
        if cho_before_update:
            self._hessian_sparsity = "diag"
        super().__init__(learning_rate, hook, max_percentage_update, bound_consistency)
        self.weight_decay = weight_decay
        self.cho_before_update = cho_before_update
        self.cho_maxiter = cho_maxiter
        if cho_solve_kwargs is None:
            cho_solve_kwargs = {"check_finite": False}
        self.cho_solve_kwargs = cho_solve_kwargs

    def _second_order_update(
        self, gradient: npt.NDArray[np.floating], hessian: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value
        lr = self.lr_scheduler.value
        w = self.weight_decay
        cho_kw = self.cho_solve_kwargs
        L = cholesky_added_multiple_identity(hessian, maxiter=self.cho_maxiter)

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            dtheta = _nm_unconstrained(theta, gradient, L, lr, w, cho_kw)
            return theta + dtheta, None
        H, G = _nm_constrained(
            theta, gradient, hessian, L, lr, w, self.cho_before_update, cho_kw
        )
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=H, g=G, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _nm_unconstrained(
    theta: np.ndarray,
    g: np.ndarray,
    L: np.ndarray,
    lr: LrType,
    weight_decay: float,
    cho_solve_kwargs: dict,
) -> np.ndarray:
    """Computes the update's change according to the Netwon's Method."""
    if weight_decay <= 0.0:
        return -lr * cho_solve((L, True), g, **cho_solve_kwargs)
    return -np.linalg.solve(
        L @ L.T + weight_decay * np.eye(theta.shape[0]), lr * g + weight_decay * theta
    )


def _nm_constrained(
    theta: np.ndarray,
    g: np.ndarray,
    H: np.ndarray,
    L: np.ndarray,
    lr: LrType,
    weight_decay: float,
    cho_before_update: bool,
    cho_solve_kwargs: dict,
) -> tuple[np.ndarray, np.ndarray]:
    """Computes the update's change according to the Netwon's Method, which is solved
    numerically due to the presence of constraints."""
    if weight_decay <= 0.0:
        if cho_before_update:
            G = cho_solve((L, True), lr * g, **cho_solve_kwargs)
            H = np.eye(theta.shape[0])
        else:
            G = lr * g
            H = L @ L.T
    else:
        G = lr * g + weight_decay * theta
        H = L @ L.T + weight_decay * np.eye(theta.shape[0])
        if cho_before_update:
            G = np.linalg.solve(H, G)
            H = np.eye(theta.shape[0])
    return H, G



import numpy as np
import numpy.typing as npt

from ...agents.common.learning_agent import ExpType, LearningAgent
from .wrapper import LearningWrapper, SymType


class RecordUpdates(LearningWrapper[SymType, ExpType]):
    """Wrapper for recording the history of updated parametrizations by the learning
    agent.

    In other words, it records the new value of the parameters in
    :attr:`mpcrl.LearningAgent.learnable_parameters` after every call to
    :meth:`mpcrl.LearningAgent.update`. This information can be retrieved from the
    attribute :attr:`updates_history`.

    Parameters
    ----------
    agent : LearningAgent or subclass
        The agent whose updates need recording.
    """

    def __init__(self, agent: LearningAgent[SymType, ExpType]) -> None:
        super().__init__(agent)
        self.updates_history: dict[str, list[npt.NDArray[np.floating]]] = {
            p.name: [p.value] for p in agent.learnable_parameters.values()
        }

    def _on_update(self, *_, **__) -> None:
        for par in self.agent.learnable_parameters.values():
            self.updates_history[par.name].append(par.value)

    def _establish_callback_hooks(self) -> None:
        super()._establish_callback_hooks()
        # connect the agent's on_update callback to this wrapper storing action
        self._hook_callback(repr(self), "on_update", self._on_update)



"""The :mod:`mpcrl.core` submodule implements various core functionalities of the MPC-RL
framework. These allow to define various hyperparameters to, e.g., specify the frequency
of the RL updates, use experience replay, introduce exploration in the MPC policy. At
the same time, it contains foundations for specifying the learnable parameters, and
implement the hook-callback mechanism that allows to customize the training loop to a
high degree."""



"""Wrappers to record information about the environment.

This submodule contains wrappers to record information about call arguments to and
return arguments from environment's :func:`gymnasium.Env.reset` and
:func:`gymnasium.Env.step`. In particular

- the :class:`MonitorInfos` wrapper records the information dictionaries returned by the
  environment

- the :class:`MonitorEpisodes` wrapper records the observations, actions, rewards,
  episode lengths, and execution times of each episode.
"""

__all__ = ["MonitorEpisodes", "MonitorInfos"]

from .monitor_episodes import MonitorEpisodes
from .monitor_infos import MonitorInfos



import logging
import os
import pickle
import sys
import unittest
from operator import neg
from sys import platform
from warnings import filterwarnings

from csnlp.multistart import RandomStartPoint, RandomStartPoints

sys.path.append(os.path.join(os.getcwd(), "examples"))

import casadi as cs
import numpy as np
import torch
from bayesopt import BoTorchOptimizer, CstrEnv, NoisyFilterObservation, get_cstr_mpc
from dpg import LinearMpc as DpgLinearMpc
from dpg import LtiSystem as DpgLtiSystem
from gymnasium.wrappers import TimeLimit, TransformReward
from parameterized import parameterized
from q_learning import LinearMpc as QLearningLinearMpc
from q_learning import LtiSystem as QLearningLtiSystem
from q_learning_offpolicy import LinearMpc as QLearningOffPolicyLinearMpc
from q_learning_offpolicy import LtiSystem as QLearningOffPolicyLtiSystem
from q_learning_offpolicy import get_rollout_generator
from scipy.io import loadmat

from mpcrl import (
    GlobOptLearningAgent,
    LearnableParameter,
    LearnableParametersDict,
    LstdDpgAgent,
    LstdQLearningAgent,
    UpdateStrategy,
    WarmStartStrategy,
)
from mpcrl import exploration as E
from mpcrl.optim import GradientDescent, NetwonMethod
from mpcrl.wrappers.agents import Evaluate, Log, RecordUpdates
from mpcrl.wrappers.envs import MonitorEpisodes

torch.set_default_device("cpu")
torch.set_default_dtype(torch.float64)
filterwarnings("ignore", "Mpc failure", module="mpcrl")

DATA = loadmat(f"tests/data_test_examples_{platform}.mat", squeeze_me=True)


class TestExamples(unittest.TestCase):
    @parameterized.expand([(False,), (True,)])
    def test_q_learning__with_copy_and_pickle(self, use_copy: bool):
        mpc = QLearningLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        env = MonitorEpisodes(TimeLimit(QLearningLtiSystem(), max_episode_steps=100))
        agent = Log(
            RecordUpdates(
                LstdQLearningAgent(
                    update_strategy=1,
                    mpc=mpc,
                    learnable_parameters=learnable_pars,
                    discount_factor=mpc.discount_factor,
                    optimizer=NetwonMethod(
                        learning_rate=5e-2,
                        max_percentage_update=1e3,  # does nothing; allows to test qp
                    ),
                    hessian_type="approx",
                    record_td_errors=True,
                )
            ),
            level=logging.DEBUG,
            log_frequencies={"on_timestep_end": 100},
        )

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        J = agent.train(env=env, episodes=1, seed=69).item()
        agent = pickle.loads(pickle.dumps(agent))

        X = env.observations[0].squeeze().T
        U = env.actions[0].squeeze()
        R = env.rewards[0]
        TD = np.squeeze(agent.td_errors)
        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )

        # from scipy.io import savemat
        # DATA.update({
        #     "ql_J": J,
        #     "ql_X": X,
        #     "ql_U": U,
        #     "ql_R": R,
        #     "ql_TD": TD,
        #     "ql_pars": PARS,
        # })
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["ql_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(X, DATA["ql_X"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(U, DATA["ql_U"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(R, DATA["ql_R"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(TD, DATA["ql_TD"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(PARS, DATA["ql_pars"], rtol=1e0, atol=1e0)

    @parameterized.expand([(False,), (True,)])
    def test_dpg__with_copy_and_pickle(self, use_copy: bool):
        mpc = DpgLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        env = MonitorEpisodes(TimeLimit(DpgLtiSystem(), max_episode_steps=200))
        rollout_length = 50
        agent = Log(
            RecordUpdates(
                LstdDpgAgent(
                    mpc=mpc,
                    learnable_parameters=learnable_pars,
                    discount_factor=mpc.discount_factor,
                    optimizer=GradientDescent(
                        learning_rate=1e-7,
                        max_percentage_update=1e3,  # does nothing; allows to test qp
                    ),
                    update_strategy=UpdateStrategy(rollout_length, "on_timestep_end"),
                    rollout_length=rollout_length,
                    exploration=E.GreedyExploration(0.05),
                    record_policy_performance=True,
                    record_policy_gradient=True,
                )
            ),
            level=logging.DEBUG,
            log_frequencies={"on_timestep_end": 200},
        )

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        J = agent.train(env=env, episodes=1, seed=69).item()
        agent = pickle.loads(pickle.dumps(agent))

        X = env.observations[0].squeeze().T
        U = env.actions[0].squeeze()
        R = env.rewards[0]
        Jest = np.asarray(agent.policy_performances)
        Gest = np.asarray(agent.policy_gradients)
        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )

        # from scipy.io import savemat
        # DATA.update({
        #     "dpg_J": J,
        #     "dpg_X": X,
        #     "dpg_U": U,
        #     "dpg_R": R,
        #     "dpg_Jest": Jest,
        #     "dpg_Gest": Gest,
        #     "dpg_pars": PARS,
        # })
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["dpg_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(X, DATA["dpg_X"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(U, DATA["dpg_U"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(R, DATA["dpg_R"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(Jest, DATA["dpg_Jest"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(Gest, DATA["dpg_Gest"], rtol=1e3, atol=1e3)
        np.testing.assert_allclose(PARS, DATA["dpg_pars"], rtol=1e0, atol=1e0)

    @parameterized.expand([(False,), (True,)])
    def test_bayesopt__with_copy_and_pickle(self, use_copy: bool):
        torch.manual_seed(0)
        env = MonitorEpisodes(TimeLimit(CstrEnv(4e3), max_episode_steps=10))
        env = TransformReward(env, neg)
        env = NoisyFilterObservation(env, [1, 2])
        multistarts = 4
        mpc = get_cstr_mpc(env, horizon=7, multistarts=multistarts, n_jobs=multistarts)
        pars = mpc.parameters
        Y = mpc.variables["y"].shape
        U = mpc.variables["u"].shape
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(n, pars[n].shape, (ub + lb) / 2, lb, ub, pars[n])
                for n, lb, ub in [("narx_weights", -2, 2), ("backoff", 0, 5)]
            )
        )
        warmstart = WarmStartStrategy(
            random_points=RandomStartPoints(
                {
                    "y": RandomStartPoint("normal", scale=[[1.0], [20.0]], size=Y),
                    "u": RandomStartPoint("normal", scale=5.0, size=U),
                },
                biases={
                    "y": CstrEnv.x0[[1, 2]].reshape(-1, 1),
                    "u": sum(CstrEnv.inflow_bound) / 2,
                },
                multistarts=multistarts,
            ),
        )
        agent = GlobOptLearningAgent(
            mpc=mpc,
            learnable_parameters=learnable_pars,
            optimizer=BoTorchOptimizer(initial_random=2, seed=42),
            warmstart=warmstart,
        )
        agent = RecordUpdates(agent)
        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy

        J = agent.train(env=env, episodes=6, seed=69, raises=False)
        agent = pickle.loads(pickle.dumps(agent))

        X = np.squeeze(env.get_wrapper_attr("observations"))
        U = np.squeeze(env.get_wrapper_attr("actions"), (2, 3))
        R = np.squeeze(env.get_wrapper_attr("rewards"))

        # from scipy.io import savemat
        # DATA.update({"bo_J": J, "bo_X": X, "bo_U": U, "bo_R": R})
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["bo_J"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(X, DATA["bo_X"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(U, DATA["bo_U"], rtol=1e-2, atol=1e-2)
        np.testing.assert_allclose(R, DATA["bo_R"], rtol=1e-2, atol=1e-2)

    @parameterized.expand([(False,), (True,)])
    def test_q_learning_offpolicy__with_copy_and_pickle(self, use_copy: bool):
        mpc = QLearningOffPolicyLinearMpc()
        learnable_pars = LearnableParametersDict[cs.SX](
            (
                LearnableParameter(
                    name=name, shape=val.shape, value=val, sym=mpc.parameters[name]
                )
                for name, val in mpc.learnable_pars_init.items()
            )
        )
        seed = np.random.default_rng(69)
        agent = Evaluate(
            Log(
                RecordUpdates(
                    LstdQLearningAgent(
                        mpc=mpc,
                        learnable_parameters=learnable_pars,
                        discount_factor=mpc.discount_factor,
                        update_strategy=1,
                        optimizer=NetwonMethod(learning_rate=5e-2),
                        hessian_type="approx",
                        record_td_errors=True,
                        remove_bounds_on_initial_action=True,
                    )
                ),
                level=logging.DEBUG,
                log_frequencies={"on_episode_end": 1},
            ),
            eval_env=TimeLimit(QLearningOffPolicyLtiSystem(), 100),
            hook="on_episode_end",
            frequency=2,
            n_eval_episodes=2,
            seed=seed,
        )
        generate_rollout = get_rollout_generator(rollout_seed=69)

        agent_copy = agent.copy()
        if use_copy:
            agent = agent_copy
        agent.train_offpolicy(
            episode_rollouts=(generate_rollout(n) for n in range(6)), seed=seed
        )
        J = np.asarray(agent.eval_returns)
        agent = pickle.loads(pickle.dumps(agent))

        parnames = ["V0", "x_lb", "x_ub", "b", "f", "A", "B"]
        PARS = np.concatenate(
            [np.reshape(agent.updates_history[n], -1) for n in parnames]
        )
        TD = np.squeeze(agent.td_errors)

        # from scipy.io import savemat
        # DATA.update({"ql_offpol_J": J, "ql_offpol_TD": TD, "ql_offpol_pars": PARS})
        # savemat(f"tests/data_test_examples_{platform}.mat", DATA)

        np.testing.assert_allclose(J, DATA["ql_offpol_J"], rtol=1e0, atol=1e0)
        np.testing.assert_allclose(TD, DATA["ql_offpol_TD"], rtol=1e1, atol=1e1)
        np.testing.assert_allclose(PARS, DATA["ql_offpol_pars"], rtol=1e0, atol=1e0)


if __name__ == "__main__":
    unittest.main()



from collections import deque
from time import perf_counter
from typing import Any, Deque, Optional, SupportsFloat, TypeVar

import numpy as np
import numpy.typing as npt
from gymnasium import Env, Wrapper, utils

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class MonitorEpisodes(
    Wrapper[ObsType, ActType, ObsType, ActType], utils.RecordConstructorArgs
):
    """This wrapper keeps track of observations, actions, rewards, episode lengths, and
    execution times of each episode.

    These are saved in the following fields:
     - observations (:attr:`observations`)
     - actions (:attr:`actions`)
     - costs/rewards (:attr:`rewards`)
     - episode length (:attr:`MonitorEpisodes.episode_lengths`)
     - episode execution time (:attr:`exec_times`)

    that the environment is subject to during the learning process. Note that these are
    effectively saved in each corresponding field only when the episode is done
    (terminated or truncated). This means that if an episode, e.g., the last one, has
    not been terminated or truncated, these fields will not have recorded its data
    (which can be found in the internal attributes).

    Parameters
    ----------
    env : Env[ObsType, ActType]
        The environment to apply the wrapper to.
    deque_size : int, optional
        The maximum number of episodes to hold as historical data in the internal
        deques. By default, `None`, i.e., unlimited.

    Examples
    --------
    After the completion of an episode, these fields will look like this:

    >>> env.observations = <deque of each episode's observations>
    ... env.actions = <deque of each episode's actions>
    ... env.rewards = <deque of each episode's rewards>
    ... env.episode_lengths = <deque of each episode's episode length>
    ... env.exec_times = <deque of each episode's execution time>
    """

    def __init__(
        self, env: Env[ObsType, ActType], deque_size: Optional[int] = None
    ) -> None:
        utils.RecordConstructorArgs.__init__(self, deque_size=deque_size)
        Wrapper.__init__(self, env)
        # long-term storages
        self.observations: Deque[npt.NDArray[ObsType]] = deque(maxlen=deque_size)
        self.actions: Deque[npt.NDArray[ActType]] = deque(maxlen=deque_size)
        self.rewards: Deque[npt.NDArray[np.floating]] = deque(maxlen=deque_size)
        self.episode_lengths: Deque[int] = deque(maxlen=deque_size)
        self.exec_times: Deque[float] = deque(maxlen=deque_size)
        # current-episode-storages
        self.ep_observations: list[ObsType] = []
        self.ep_actions: list[ActType] = []
        self.ep_rewards: list[SupportsFloat] = []
        self.t0: float = perf_counter()
        self.ep_length: int = 0

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
    ) -> tuple[ObsType, dict[str, Any]]:
        observation, info = super().reset(seed=seed, options=options)
        self._clear_ep_data()
        self.ep_observations.append(observation)
        return observation, info

    def step(
        self, action: ActType
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        obs, reward, terminated, truncated, info = super().step(action)

        # accumulate data
        self.ep_observations.append(obs)
        self.ep_actions.append(action)
        self.ep_rewards.append(reward)
        self.ep_length += 1

        # if episode is done, save the current data to history
        if terminated or truncated:
            # append data
            self.observations.append(np.asarray(self.ep_observations))
            self.actions.append(np.asarray(self.ep_actions))
            self.rewards.append(np.asarray(self.ep_rewards))
            self.episode_lengths.append(self.ep_length)
            self.exec_times.append(perf_counter() - self.t0)

            # clear this episode's data
            self._clear_ep_data()

        return obs, reward, terminated, truncated, info

    def _clear_ep_data(self) -> None:
        # clear this episode's lists and reset counters
        self.ep_observations.clear()
        self.ep_actions.clear()
        self.ep_rewards.clear()
        self.t0 = perf_counter()
        self.ep_length = 0



import pickle
import unittest
from functools import lru_cache
from itertools import product
from unittest.mock import MagicMock, Mock, call
from warnings import catch_warnings

import casadi as cs
import numpy as np
from csnlp import Nlp, scaling
from csnlp.core.solutions import EagerSolution
from csnlp.multistart import StackedMultistartNlp
from csnlp.wrappers import Mpc, NlpScaling
from parameterized import parameterized, parameterized_class
from scipy import io as matio

from mpcrl import (
    Agent,
    ExperienceReplay,
    LearnableParametersDict,
    LearningAgent,
    MpcSolverWarning,
)
from mpcrl import exploration as E
from mpcrl import schedulers as S
from mpcrl.util.seeding import mk_seed

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}
RESULTS = matio.loadmat(r"tests/data_test_agent.mat")


@lru_cache
def get_dynamics(g: float, alpha: float, dt: float) -> cs.Function:
    x, u, d = cs.MX.sym("x", 3), cs.MX.sym("u", 2), cs.MX.sym("d", 3)
    x_next = x + cs.vertcat(x[1], u[0] / x[2] - g, -alpha * u[0]) * dt + d * 0
    return cs.Function("F", [x, u, d], [x_next], ["x", "u", "d"], ["x+"])


def get_mpc(horizon: int, multistart: bool):
    N = horizon
    T = 100
    K = 3
    dt = T / N
    yT = 100000
    g = 9.81
    alpha = 1 / (300 * g)
    y_nom = 1e5
    v_nom = 2e3
    m_nom = 3e5
    u_nom = 1e8
    scaler = scaling.Scaler()
    scaler.register("y", scale=y_nom)
    scaler.register("y_0", scale=y_nom)
    scaler.register("v", scale=v_nom)
    scaler.register("v_0", scale=v_nom)
    scaler.register("m", scale=m_nom)
    scaler.register("m_0", scale=m_nom)
    scaler.register("u1", scale=u_nom)
    scaler.register("u2", scale=u_nom)
    nlp = (
        StackedMultistartNlp[cs.MX](sym_type="MX", starts=K)
        if multistart
        else Nlp[cs.MX](sym_type="MX")
    )
    nlp = NlpScaling[cs.MX](nlp, scaler=scaler, warns=False)
    mpc = Mpc[cs.MX](nlp, prediction_horizon=N)
    y, _ = mpc.state("y")
    _, _ = mpc.state("v")
    m, _ = mpc.state("m", lb=0)
    mpc.action("u1", lb=0, ub=5e7)
    u2, _ = mpc.action("u2", lb=0, ub=5e7)
    mpc.disturbance("d", 3)
    mpc.set_dynamics(get_dynamics(g, alpha, dt))
    mpc.constraint("yT", y[-1], "==", yT)
    mpc.minimize(m[0] - m[-1] + cs.sum2(u2))
    mpc.init_solver(OPTS)
    return mpc


class DummyLearningAgent(LearningAgent):
    def update(self, *args, **kwargs):
        return

    def train_one_episode(self, *args, **kwargs):
        return


@parameterized_class("multistart_nlp", [(True,), (False,)])
class TestAgent(unittest.TestCase):
    def test_init__raises__mpc_with_no_actions(self):
        with self.assertRaisesRegex(
            ValueError, "Expected Mpc with na>0; got na=0 instead."
        ):
            Agent(mpc=Mpc(Nlp(), 4))

    def test_init__instantiates_V_and_Q_as_two_different_mpcs(self):
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp))
        self.assertIsInstance(agent.Q, Mpc)
        self.assertIsInstance(agent.V, Mpc)
        self.assertIsNot(agent.Q, agent.V)

    @parameterized.expand([(None,), ("gradient-based",), ("additive",)])
    def test_init__instantiates_V_and_Q_correctly(self, exploration_mode: str):
        if exploration_mode is None:
            exploration = E.NoExploration()
        else:
            exploration = E.GreedyExploration(0.5, mode=exploration_mode)
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp), exploration=exploration)

        V_pars_keys = agent.V.parameters.keys()
        Q_pars_keys = agent.Q.parameters.keys()
        if exploration_mode is None or exploration_mode == "additive":
            self.assertNotIn(agent.cost_perturbation_parameter, V_pars_keys)
        else:
            self.assertIn(agent.cost_perturbation_parameter, V_pars_keys)
        self.assertNotIn(agent.cost_perturbation_parameter, Q_pars_keys)
        self.assertIn(agent.init_action_parameter, Q_pars_keys)
        self.assertIn(agent.init_action_constraint, agent.Q.constraints.keys())
        self.assertNotIn(agent.init_action_parameter, V_pars_keys)
        self.assertNotIn(agent.init_action_constraint, agent.V.constraints.keys())

    def test_init__removes_bounds_on_initial_action_in_Q_correctly(self):
        mpc = get_mpc(3, self.multistart_nlp)
        mpc.unwrapped.remove_variable_bounds = Mock()
        agent = Agent(mpc=mpc, remove_bounds_on_initial_action=True)
        self.assertFalse(agent.V.remove_variable_bounds.called)
        calls = agent.Q.remove_variable_bounds.call_args_list
        self.assertEqual(len(calls), mpc.na)
        for i, mcall in enumerate(calls):
            expected_name = f"u{i + 1}"
            na_ = mpc.actions[expected_name].size1()
            name, direction, idx = mcall.args
            self.assertEqual(name, expected_name)
            self.assertEqual(direction, "both")
            self.assertEqual(list(idx), [(r, 0) for r in range(na_)])

    def test_unwrapped(self):
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp))
        agent2 = agent.unwrapped
        self.assertIs(agent, agent2)

    @parameterized.expand([(True,), (False,)])
    def test__is_deepcopyable_and_pickleable(self, copy: bool):
        epsilon, epsilon_decay_rate = 0.7, 0.75
        strength, strength_decay_rate = 0.5, 0.75
        epsilon_scheduler = S.ExponentialScheduler(epsilon, epsilon_decay_rate)
        strength_scheduler = S.ExponentialScheduler(strength, strength_decay_rate)
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )
        agent1 = Agent(mpc=get_mpc(3, self.multistart_nlp))
        agent1._exploration = exploration  # a bit of cheating

        if copy:
            agent2 = agent1.copy()
        else:
            agent2: Agent = pickle.loads(pickle.dumps(agent1))

        self.assertIsNot(agent1, agent2)
        self.assertIsNot(agent1.Q, agent2.Q)
        self.assertIsNot(agent1.V, agent2.V)
        self.assertIs(agent1.exploration, exploration)
        exp1, exp2 = agent1.exploration, agent2.exploration
        self.assertIsNot(exp1, exp2)
        self.assertIsNot(exp1, exp2)
        for scheduler in ["epsilon_scheduler", "strength_scheduler"]:
            sc1, sc2 = getattr(exp1, scheduler), getattr(exp2, scheduler)
            self.assertIsNot(sc1, sc2)
            self.assertEqual(sc1.value, sc2.value)
            self.assertEqual(sc1.factor, sc2.factor)

    @parameterized.expand(product(["V", "Q"], [False, True], [False, True]))
    def test_solve_mpc__calls_mpc_with_correct_args(
        self, mpctype: str, vector: bool, multiple_pars: bool
    ):
        starts = 3
        horizon = 3
        multiple_pars &= self.multistart_nlp
        fixed_pars = {
            Agent.cost_perturbation_parameter: [42, 69],
            "d": cs.DM([5, 6, 7]),
        }
        fixed_pars_ = (
            (fixed_pars.copy() for _ in range(starts))
            if multiple_pars
            else fixed_pars.copy()
        )
        mpc = get_mpc(horizon, self.multistart_nlp)
        exploration = E.GreedyExploration(0.5, mode="gradient-based")
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars_, exploration=exploration)
        mpc: Mpc[cs.SX] = getattr(agent, mpctype)

        s = {"y": 0, "v": 10, "m": 5e5}
        if vector:
            s = cs.DM(s.values())
        if mpctype == "V":
            a = None
        else:
            a = {"u1": 1, "u2": 2}
            if vector:
                a = cs.DM(a.values())

        pert = cs.DM([65, 79])
        vals0 = object()
        sol = EagerSolution(
            f=None,
            p_sym=None,
            p=None,
            x_sym=None,
            x=None,
            lam_g_and_h_sym=None,
            lam_g_and_h=None,
            lam_lbx_and_ubx_sym=None,
            lam_lbx_and_ubx=None,
            vars=None,
            vals=vals0,
            dual_vars=None,
            dual_vals=None,
            stats={"success": True},
        )
        agent._last_solution = sol
        method = "solve_multi" if self.multistart_nlp else "solve"
        setattr(mpc.nlp, method, MagicMock(return_value=sol))

        agent._solve_mpc(mpc, state=s, action=a, perturbation=pert)

        call_pars = {
            **fixed_pars,
            "y_0": s[0] if vector else s["y"],
            "v_0": s[1] if vector else s["v"],
            "m_0": s[2] if vector else s["m"],
        }
        if mpctype == "V":
            call_pars[Agent.cost_perturbation_parameter] = pert
        else:
            call_pars[Agent.init_action_parameter] = a if vector else cs.DM(a.values())

        overwritten_method = getattr(mpc.nlp, method)
        overwritten_method.assert_called_once()
        call_args = overwritten_method.call_args.args
        actual_call_pars, actual_call_vals0 = call_args
        if multiple_pars:
            for pars_i in actual_call_pars:
                self.assertEqual(len(mpc.unwrapped._pars.keys() - pars_i.keys()), 0)
                for key in call_pars:
                    np.testing.assert_allclose(pars_i[key], call_pars[key], rtol=0)
        else:
            pars = actual_call_pars
            self.assertEqual(len(mpc.unwrapped._pars.keys() - pars.keys()), 0)
            for key in call_pars:
                np.testing.assert_allclose(pars[key], call_pars[key], rtol=0)
        self.assertIs(actual_call_vals0, vals0)

    @parameterized.expand(product((False, True), (False, True)))
    def test_state_value__computes_right_solution(
        self, vector: bool, deterministic: bool
    ):
        starts = 3
        horizon = 100
        fixed_pars = {"d": cs.DM([5, 6, 7])}
        if self.multistart_nlp:
            fixed_pars = [fixed_pars.copy() for _ in range(starts)]

        mpc = get_mpc(horizon, self.multistart_nlp)
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars)

        state = {"y": 0, "v": 0, "m": 5e5}
        vals0 = {"y": 0, "v": 0, "m": 5e5, "u1": 1e8, "u2": 0}
        agent._last_solution = EagerSolution(
            f=0.0,
            p_sym=None,
            p=None,
            x_sym=None,
            x=None,
            lam_g_and_h_sym=None,
            lam_g_and_h=None,
            lam_lbx_and_ubx_sym=None,
            lam_lbx_and_ubx=None,
            vars=None,
            vals=vals0,
            dual_vars=None,
            dual_vals=None,
            stats=None,
        )
        if vector:
            state = cs.DM(state.values())

        action, sol = agent.state_value(
            state=state, vals0=None, deterministic=deterministic
        )
        self.assertTrue(sol.success)
        np.testing.assert_array_equal(
            action, cs.vertcat(sol.vals["u1"][:, 0], sol.vals["u2"][:, 0])
        )
        np.testing.assert_allclose(sol.f, RESULTS["state_value_f"].item(), rtol=1e-3)
        np.testing.assert_allclose(
            sol.vals["u1"],
            RESULTS["state_value_us"],
            rtol=1e-7,
            atol=1e-7,
        )

    @parameterized.expand(product((False, True), (False, True)))
    def test_action_value__computes_right_solution(self, vector: bool, a_optimal: bool):
        starts = 3
        horizon = 100
        fixed_pars = {"d": cs.DM([5, 6, 7])}
        if self.multistart_nlp:
            fixed_pars = [fixed_pars.copy() for _ in range(starts)]

        mpc = get_mpc(horizon, self.multistart_nlp)
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars)

        a_opt, a_subopt = 5e7, 1.42e7
        state = {"y": 0, "v": 0, "m": 5e5}
        action = {"u1": a_opt if a_optimal else a_subopt, "u2": 0}
        vals0 = {**state, **action}
        agent._last_solution = EagerSolution(
            f=None,
            p_sym=None,
            p=None,
            x_sym=None,
            x=None,
            lam_g_and_h_sym=None,
            lam_g_and_h=None,
            lam_lbx_and_ubx_sym=None,
            lam_lbx_and_ubx=None,
            vars=None,
            vals=vals0,
            dual_vars=None,
            dual_vals=None,
            stats=None,
        )
        if vector:
            state = cs.DM(state.values())
            action = cs.DM(action.values())

        sol = agent.action_value(state=state, action=action, vals0=None)
        self.assertTrue(sol.success)
        u1_0_0 = sol.value(mpc.unscaled_variables["u1"][0, 0])
        if a_optimal:
            np.testing.assert_allclose(u1_0_0, a_opt)
            np.testing.assert_allclose(
                sol.vals["u1"],
                RESULTS["state_value_us"],
                rtol=1e-7,
                atol=1e-7,
            )
        else:
            np.testing.assert_allclose(u1_0_0, a_subopt)
            self.assertTrue(sol.f >= RESULTS["state_value_f"].item())

    def test_evaluate__performs_correct_calls(self):
        seed = 69
        episodes = 3
        episode_length = 10
        Ttot = episodes * episode_length
        reset_states = [object() for _ in range(episodes)]
        reset_options = object()
        step_states = [object() for _ in range(Ttot)]
        rewards = np.random.randn(Ttot)
        truncateds = [False] * Ttot
        terminateds = np.full(Ttot, 0, dtype=bool)
        terminateds[::-episode_length] = True
        infos = [{}] * Ttot
        env = MagicMock()
        env.reset = Mock(side_effect=zip(reset_states, infos))
        env.step = Mock(
            side_effect=zip(step_states, rewards, truncateds, terminateds, infos)
        )
        actions1 = np.random.randn(Ttot, 2, 1)
        actions2 = np.random.randn(Ttot, 3, 1)
        actions = [cs.vertcat(u1, u2) for u1, u2 in zip(actions1, actions2)]
        successes = [True] * (Ttot - 1) + [False]
        sols = map(
            lambda u1, u2, success: EagerSolution(
                f=None,
                p_sym=None,
                p=None,
                x_sym=None,
                x=None,
                lam_g_and_h_sym=None,
                lam_g_and_h=None,
                lam_lbx_and_ubx_sym=None,
                lam_lbx_and_ubx=None,
                vars=None,
                vals={"u1": u1, "u2": u2},
                dual_vars=None,
                dual_vals=None,
                stats={"success": success, "return_status": "bad"},
            ),
            actions1,
            actions2,
            successes,
        )
        agent = Agent(mpc=get_mpc(3, False))
        agent.state_value = Mock(side_effect=zip(actions, sols))
        deterministic = object()

        with catch_warnings(record=True) as cw:
            returns = Agent.evaluate(
                agent,
                env,
                episodes,
                deterministic=deterministic,
                seed=seed,
                raises=False,
                env_reset_options=reset_options,
            )

        np.testing.assert_allclose(returns, rewards.reshape(-1, episode_length).sum(1))
        rng = np.random.default_rng(seed)
        env.reset.assert_has_calls(
            [call(seed=mk_seed(rng), options=reset_options) for i in range(episodes)]
        )
        for mcall, u1, u2 in zip(env.step.mock_calls, actions1, actions2):
            self.assertEqual(len(mcall.args), 1)
            np.testing.assert_array_equal(mcall.args[0], cs.vertcat(u1, u2))
        states = np.column_stack(
            (reset_states, np.asarray(step_states).reshape(-1, episode_length))
        )[:, :-1].flatten()
        agent.state_value.assert_has_calls(
            [call(state, deterministic) for state in states]
        )
        self.assertEqual(len(cw), 1)
        self.assertIs(cw[0].category, MpcSolverWarning)


class TestLearningAgent(unittest.TestCase):
    def test_init__creates_default_experience_and_exploration(self):
        learnable_parameters = LearnableParametersDict()
        agent = DummyLearningAgent(
            mpc=get_mpc(3, False),
            learnable_parameters=learnable_parameters,
            update_strategy=1,
        )
        self.assertIsInstance(agent.exploration, E.ExplorationStrategy)
        self.assertIsInstance(agent.experience, ExperienceReplay)
        self.assertIs(agent.learnable_parameters, learnable_parameters)

    def test_store_experience__appends_experience(self):
        experience = MagicMock()
        experience.append = Mock()
        item = object()
        agent = DummyLearningAgent(
            mpc=get_mpc(3, False),
            learnable_parameters={},
            experience=experience,
            update_strategy=1,
        )
        self.assertIs(agent.experience, experience)
        agent.store_experience(item)
        experience.append.assert_called_once_with(item)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from collections.abc import Collection, Iterable
from typing import Any, Callable, Generic, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from gymnasium import Env
from gymnasium.spaces import Box

from ...core.callbacks import LearningAgentCallbackMixin
from ...core.experience import ExperienceReplay
from ...core.parameters import LearnableParametersDict
from ...core.update import UpdateStrategy
from ...util.seeding import RngType, mk_seed
from .agent import ActType, Agent, ObsType, SymType, _update_dicts

ExpType = TypeVar("ExpType")


class LearningAgent(
    Agent[SymType], LearningAgentCallbackMixin, ABC, Generic[SymType, ExpType]
):
    r"""Base abstract class for a learning agent with MPC as policy provider. The main
    method :meth:`update`, which is called to update the learnable parameters of the MPC
    according to the underlying learning methodology (e.g., Bayesian Optimization, RL,
    etc.), is abstract and must be implemented by inheriting classes based on their
    learning algorithm.

    Aside of :meth:`update`, this class also provides also the basic skeleton for both
    on-policy and off-policy learning, which require further ad-hoc implementations. For
    on-policy learning, the :meth:`train` method should be run, which requires the
    implementation of :meth:`train_one_episode`. For off-policy learning, run
    :meth:`train_offpolicy` and implement :meth:`train_one_rollout` instead. At least
    one of the two implementations is required in the inheriting class, depending on the
    learning algorithm. Some algorithms might support both.

    Parameters
    ----------
    update_strategy : UpdateStrategy or int
        The strategy used to decide which frequency to update the mpc parameters with.
        If an ``int`` is passed, then the default strategy that updates every ``n``
        env's steps is used (where ``n`` is the argument passed); otherwise, an instance
        of :class:`core.update.UpdateStrategy` can be passed to specify the
        desired strategy in more details.
    learnable_parameters : :class:`core.parameters.LearnableParametersDict`
        A special dict containing the learnable parameters of the MPC (usually referred
        to as :math:`\theta`), together with their bounds and values. This dict is
        complementary to :attr:`fixed_parameters`, which contains the MPC parameters
        that are not learnt by the agent.
    experience : int or ExperienceReplay, optional
        The container for experience replay memory. If ``None`` is passed, then a memory
        with unitary length is created, i.e., it keeps only the latest memory
        transition. If an integer ``n`` is passed, then a memory with the length ``n``
        is created and with sample size ``n``. Otherwise, pass an instance of
        :class:`core.experience.ExperienceReplay` to specify the requirements in more
        details.
    kwargs
        Additional arguments to be passed to :class:`Agent`.

    Notes
    -----
    This class makes no assumptions on the learning methodology used to update the MPC's
    learnable parameters. This could be either gradient-based or gradient-free, but the
    logic implemented in this class should be largely remain untouched.
    """

    def __init__(
        self,
        update_strategy: Union[int, UpdateStrategy],
        learnable_parameters: LearnableParametersDict[SymType],
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        **kwargs: Any,
    ) -> None:
        Agent.__init__(self, **kwargs)
        LearningAgentCallbackMixin.__init__(self)
        self._learnable_pars = learnable_parameters
        if experience is None:
            experience = ExperienceReplay(maxlen=1)
        elif isinstance(experience, int):
            experience = ExperienceReplay(maxlen=experience, sample_size=experience)
        self._experience = experience
        if not isinstance(update_strategy, UpdateStrategy):
            update_strategy = UpdateStrategy(update_strategy, "on_timestep_end")
        self._update_strategy = update_strategy
        self._raises: bool
        self._is_training = False
        self._establish_callback_hooks()

    @property
    def experience(self) -> ExperienceReplay[ExpType]:
        """Gets the experience replay memory of the agent."""
        return self._experience

    @property
    def update_strategy(self) -> UpdateStrategy:
        """Gets the update strategy of the agent."""
        return self._update_strategy

    @property
    def learnable_parameters(self) -> LearnableParametersDict[SymType]:
        """Gets the parameters of the MPC that can be learnt by the agent."""
        return self._learnable_pars

    def reset(self, seed: RngType = None) -> None:
        super().reset(seed)
        self.experience.reset(seed)

    def store_experience(self, item: ExpType) -> None:
        """Stores the given item in the agent's :attr:`experience` for later usage in
        updating the parametrization.

        Parameters
        ----------
        item : ExpType
            Item to be stored in memory.
        """
        self._experience.append(item)

    def evaluate(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        self._is_training = False
        return super().evaluate(*args, **kwargs)

    def train(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
    ) -> npt.NDArray[np.floating]:
        """On-policy training of the agent on an environment.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            The gym environment where to train the agent on.
        episodes : int
            Number of training episodes.
        seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
            Seed for the agent's and env's random number generator. By default ``None``.
        raises : bool, optional
            If ``True``, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns for each training episode.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on ``raises``) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on ``raises``) if the update
            fails.
        """
        if hasattr(env, "action_space"):
            assert isinstance(env.action_space, Box), "Env action space must be a Box,"
        rng = np.random.default_rng(seed)
        self.reset(rng)
        self._is_training = True
        self._raises = raises
        returns = np.zeros(episodes, float)

        self.on_training_start(env)
        for episode in range(episodes):
            state, _ = env.reset(seed=mk_seed(rng), options=env_reset_options)
            self.on_episode_start(env, episode, state)
            r = self.train_one_episode(env, episode, state, raises)
            self.on_episode_end(env, episode, r)
            returns[episode] = r

        self.on_training_end(env, returns)
        return returns

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        """On-policy training of the agent on an environment for one single episode.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            The gym environment where to train the agent on.
        episode : int
            Number of the current training episode.
        init_state : observation type
            Initial state/observation of the environment.
        raises : bool, optional
            If ``True``, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Returns
        -------
        float
            The cumulative rewards for this training episode.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on ``raises``) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on ``raises``) if the update
            fails.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `train_one_episode` for "
            "on-policy learning."
        )

    def train_offpolicy(
        self,
        episode_rollouts: Iterable[Iterable[Any]],
        seed: RngType = None,
        raises: bool = True,
    ) -> None:
        """Off-policy training of the agent on an environment.

        Parameters
        ----------
        episode_rollouts : iterable of iterables of any
            An iterable of episodical rollouts generated in an
            off-policy fashion. Each rollout is itself a sequence of transitions, e.g.,
            SARSA tuples. In other words, `episode_rollouts` is a sequence of sequences
            of tuples. However, in general, its nature and the tuples' can widely differ
            from learning algorithm to learning algorithm.
        seed : None, int, array_like of ints, SeedSequence, BitGenerator, Generator
            Seed for the agent's random number generator. By default ``None``.
        raises : bool, optional
            If ``True``, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on ``raises``) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on ``raises``) if the update
            fails.
        """
        rng = np.random.default_rng(seed)
        self.reset(rng)
        self._raises = raises
        env_proxy = "off-policy"

        self._is_training = True
        self.on_training_start(env_proxy)
        for episode, rollout in enumerate(episode_rollouts):
            self.on_episode_start(env_proxy, episode, float("nan"))
            self.train_one_rollout(rollout, episode, raises)
            self.on_episode_end(env_proxy, episode, float("nan"))

        self.on_training_end(env_proxy, np.empty(0))

    def train_one_rollout(
        self, rollout: Iterable[Any], episode: int, raises: bool = True
    ) -> None:
        """Train the agent in an off-policy manner on the given rollout.

        Parameters
        ----------
        rollout : iterable of any
            Rollout, i.e., a sequence of transitions generated off-policy, e.g., SARSA
            tuples.  However, in general, these tuples can be of different nature,
            depending on the specific learning algorithm.
        raises : bool, optional
            If ``True``, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on ``raises``) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on ``raises``) if the update
            fails.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `train_offpolicy` for "
            "off-policy learning."
        )

    @abstractmethod
    def update(self) -> Optional[str]:
        r"""Updates the learnable parameters (usually referred to as :math:`\theta`) of
        the MPC according to the agent's learning algorithm.

        Returns
        -------
        errormsg : str or None
            In case the update fails, an error message is returned to be raised as error
            or warning; otherwise, `None` is returned.
        """

    def _establish_callback_hooks(self) -> None:
        super()._establish_callback_hooks()
        # hook exploration (only if necessary)
        exploration_hook = self._exploration.hook
        if exploration_hook is not None:
            self._hook_callback(
                repr(self._exploration), exploration_hook, self._exploration.step
            )
        # hook updates (always necessary)
        update_hook = self._update_strategy.hook
        func: Callable[..., None] = (
            (lambda _, e, __: self._check_and_perform_update(e, None))
            if update_hook == "on_episode_end"
            else (lambda _, e, t: self._check_and_perform_update(e, t))
        )
        self._hook_callback(repr(self._update_strategy), update_hook, func)

    def _check_and_perform_update(self, episode: int, timestep: Optional[int]) -> None:
        """Internal utility to check if an update is due and perform it."""
        if not self._is_training or not self._update_strategy.can_update():
            return
        update_msg = self.update()
        if update_msg is not None:
            self.on_update_failure(episode, timestep, update_msg, self._raises)
        self.on_update()

    def _get_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Internal utility to retrieve parameters of the MPC in order to solve it.
        :class:`LearningAgent` returns both fixed and learnable parameters."""
        learnable_pars = self._learnable_pars.value_as_dict
        fixed_pars = self._fixed_pars
        if fixed_pars is None:
            return learnable_pars
        if isinstance(fixed_pars, dict):
            fixed_pars.update(learnable_pars)
            return fixed_pars
        return tuple(_update_dicts(self._fixed_pars, learnable_pars))



from collections import deque
from collections.abc import Iterable
from logging import warn
from typing import Any, Deque, Optional, SupportsFloat, TypeVar

from gymnasium import Env, Wrapper, utils

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


def _compact_dicts(
    dicts: Iterable[dict[str, Any]], fill_value: Any = None
) -> dict[str, list[Any]]:
    """Compacts an iterable of dictionaries into a single dict with lists of entries. If
    an entry is missing for any given dict, ``fill_value`` is used in place.

    Parameters
    ----------
    dicts : iterable of dicts of (str, any)
        Dictionaries to be compacted into a single one.
    fill_value : any, optional
        The value to be used to fill in missing values.

    Returns
    -------
    dict of (str, list of any | fill_value)
        A unique dictionary made from all the passed dicts.
    """
    out: dict[str, list[Any]] = {}
    for i, dict_ in enumerate(dicts):
        for k, v in dict_.items():
            if k in out:
                out[k].append(v)
            else:
                out[k] = [fill_value] * i + [v]
        for k in out.keys() - dict_.keys():
            out[k].append(fill_value)
    return out


class MonitorInfos(
    Wrapper[ObsType, ActType, ObsType, ActType], utils.RecordConstructorArgs
):
    """This wrapper keeps track of the infos that are generated by calls to
    :func:`gymnasium.Env.reset` and :func:`gymnasium.Env.step`.

    Parameters
    ----------
    env : Env[ObsType, ActType]
        The environment to apply the wrapper to.
    deque_size : int, optional
        The maximum number of episodes to hold as historical data in the internal
        deques. By default, ``None``, i.e., unlimited.
    """

    def __init__(
        self, env: Env[ObsType, ActType], deque_size: Optional[int] = None
    ) -> None:
        utils.RecordConstructorArgs.__init__(self, deque_size=deque_size)
        Wrapper.__init__(self, env)
        # long-term storages
        self.reset_infos: Deque[dict[str, Any]] = deque(maxlen=deque_size)
        self.step_infos: Deque[list[dict[str, Any]]] = deque(maxlen=deque_size)
        # current-episode storages
        self.ep_step_infos: list[dict[str, Any]] = []

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
    ) -> tuple[ObsType, dict[str, Any]]:
        observation, info = super().reset(seed=seed, options=options)
        self.ep_step_infos.clear()
        self.reset_infos.append(info)
        return observation, info

    def step(
        self, action: ActType
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        obs, reward, terminated, truncated, info = super().step(action)
        self.ep_step_infos.append(info)
        if terminated or truncated:
            self.step_infos.append(self.ep_step_infos.copy())
            self.ep_step_infos.clear()
        return obs, reward, terminated, truncated, info

    def finalized_reset_infos(self, fill_value: Any = None) -> dict[str, list[Any]]:
        """Returns a compacted final dictionary containing the reset infos. Missing
        values are filled automatically.

        Parameters
        ----------
        fill_value : Any, optional
            The value to be used to fill in missing values.

        Returns
        -------
        dict of (str, list)
            A unique dict containing for each entry returned in the reset info the
            corresponding list of entries, one per each reset call.
        """
        return _compact_dicts(self.reset_infos, fill_value)

    def finalized_step_infos(
        self, fill_value: Any = None
    ) -> dict[str, list[list[Any]]]:
        """Returns a compacted final dictionary containing the step infos. Missing
        values are filled automatically.

        Parameters
        ----------
        fill_value : Any, optional
            The value to be used to fill in missing values.

        Returns
        -------
        dict of (str, list of lists)
            A unique dict containing for each entry returned in the step info the
            corresponding list of dicts per episode, with one entry per each step call.
        """

        if self.ep_step_infos:
            warn(
                "Internal buffer of step infos not empty, meaning that the last "
                "episode did not terminate properly.",
                RuntimeWarning,
            )
        return _compact_dicts(
            (_compact_dicts(d, fill_value) for d in self.step_infos), fill_value
        )



"""A submodule with utility functions for creating infinite iterators."""

import itertools
from collections.abc import Iterator as _Iterator


def bool_cycle(frequency: int, starts_with: bool = False) -> _Iterator[bool]:
    """Creates an infinite iterator which cycles via boolean values, where ``True``
    appears with the given frequency ``frequency``.

    Parameters
    ----------
    frequency : int
        A positive int specifing the frequency at which ``True`` appears.
    starts_with : bool, optional
        Whether the first value should be ``True`` or ``False``. By default ``False``.

    Returns
    -------
    Iterator of bool
        An iterator with the given frequency of ``True``.
    """
    iterator = itertools.cycle(
        itertools.chain(itertools.repeat(False, frequency - 1), (True,))
    )
    if starts_with:
        iterator = itertools.chain((True,), iterator)
    return iterator



"""A collection of functions for mathematical operations and utilities. In particular,
these functions support the creation of monomial basis functions for approximating the
value function, and the modifications of Hessian matrices to positive-definite ones."""

from itertools import combinations as _combinations
from typing import Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp.util.math import prod as _prod
from scipy.special import comb as _comb


def summarize_array(
    a: npt.NDArray[np.floating],
    precision: int = 3,
    ddof: int = 0,
    axis: Optional[int] = None,
) -> str:
    """Summarizes the stats of a given array, i.e., the shape, the minimum and maximum,
    and the mean and standard deviation.

    Parameters
    ----------
    a : array
        The numerical array to summarize.
    precision : int, optional
        The decimal precision, by default ``3``.
    ddof : int, optional
        Degrees of freedom used to compute the standard deviation, by default ``0``.
    axis : int, optional
        The axis along which to summarize the array, by default ``None``.

    Returns
    -------
    str
        The summarizing string.
    """
    s = a.shape
    mean = a.mean(axis=axis)
    std = a.std(ddof=ddof, axis=axis)
    min = a.min(axis=axis)
    max = a.max(axis=axis)
    return (
        f"s={s} x∈[{min:.{precision}f}, {max:.{precision}f}] "
        f"μ={mean:.{precision}f} σ={std:.{precision}f}"
    )


def cholesky_added_multiple_identity(
    A: npt.NDArray[np.floating], beta: float = 1e-3, maxiter: int = 1000
) -> npt.NDArray[np.floating]:
    r"""Lower Cholesky factorization with added multiple of the identity matrix to ensure
    positive-definitiveness from Algorithm 3.3 in :cite:`nocedal_numerical_2006`.

    The basic idea is to add a multiple of the identity matrix to the original matrix
    unitl the factorization is successful, i.e., find :math:`\tau \ge 0` such that
    :math:`L L^\top = A + \tau I` is successful.

    Parameters
    ----------
    A : array of double
        The 2D matrix to compute the cholesky factorization of.
    beta : float, optional
        Initial tolerance of the algorithm, by default ``1e-3``.
    maxiter : int, optional
        Maximum iterations of the algorithm, by default ``1000``.

    Returns
    -------
    array of double
        The lower cholesky factorization of the modified ``A`` (with the addition of
        identity matrices to ensure that it is positive-definite).

    Raises
    ------
    ValueError
        If the factorization is unsuccessful for the maximum number of iterations.
    """
    a_min = np.diag(A).min()
    tau = 0 if a_min > 0 else -a_min + beta
    identity = np.eye(A.shape[0])
    for _ in range(maxiter):
        try:
            return np.linalg.cholesky(A + tau * identity)
        except np.linalg.LinAlgError:
            tau = max(1.05 * tau, beta)
    raise ValueError("Maximum iterations reached.")


def nchoosek(n: Union[int, npt.ArrayLike], k: int) -> Union[int, np.ndarray]:
    """Emulates
    `MATLAB's nchoosek <https://www.mathworks.com/help/matlab/ref/nchoosek.html>`_, and
    returns the binomial coefficient, i.e.,  the number of combinations of ``n`` items
    taken ``k`` at a time. If ``n`` is an array, then it is flatten and all possible
    combinations of its elements are returned.

    Parameters
    ----------
    n : int or array_like
        Number of elements or array of elements to choose from.
    k : int
        Number of elements to choose.

    Returns
    -------
    int or array
        Depending on the type of input ``n``, the output is either the total number of
        combinations or the combinations in a matrix.
    """
    return (
        _comb(n, k, exact=True)
        if isinstance(n, int)
        else np.row_stack(list(_combinations(np.asarray(n).reshape(-1), k)))
    )


def monomial_powers(n: int, k: int) -> npt.NDArray[np.int_]:
    r"""Computes the powers of all ``n``-dimensional monomials of degree ``k``. In
    mathematical terms, consider the monomial
    :math:`\prod_{i=1}^n x_i^{p_i}`. Then, this functions returns all possible
    combinations of powers :math:`(p_1, p_2, \ldots, p_n)` such that their sum is equal
    to ``k``, i.e., :math:`\sum_{i=1}^n p_i = k`.

    Parameters
    ----------
    n : int
        The number of monomial elements.
    k : int
        The degree of each monomial.

    Returns
    -------
    2d array of ints
        A 2D array containing in each row the power of each index in order to obtain the
        desired monomial of power ``k``. The shape of the array is :math:`(c, n)`, where
        :math:`c = \frac{(k + n - 1)!}{k!(n - 1)!}`.

    Examples
    --------
    >>> from mpcrl.util.math import monomial_powers
    >>> monomial_powers(3, 2)
    array([[2, 0, 0],
           [1, 1, 0],
           [1, 0, 1],
           [0, 2, 0],
           [0, 1, 1],
           [0, 0, 2]])
    """
    # see https://en.wikipedia.org/wiki/Homogeneous_polynomial#Properties and
    # https://math.stackexchange.com/a/36251
    m = nchoosek(k + n - 1, n - 1)
    dividers = np.column_stack(
        (
            np.zeros((m, 1), int),
            np.row_stack(nchoosek(np.arange(1, k + n), n - 1)),
            np.full((m, 1), k + n, int),
        )
    )
    return np.flipud(np.diff(dividers, axis=1) - 1).astype(int)


def monomials_basis_function(n: int, mindegree: int, maxdegree: int) -> cs.Function:
    r"""Creates a :class:`casadi.Function` made of monomials as bases in ``n`` variables
    and degrees from ``mindegree`` to ``maxdegree``.

    In mathematical terms, given an input vector :math:`s \in \mathbb{R^n}` and the
    minimum and maximum degrees :math:`m` and :math:`M`, consider the degree
    :math:`d_i=m+i` with :math:`i=0,\ldots,M-m`. Associated to this degree, there exist
    :math:`c_i = \frac{(d_i + n - 1)!}{d_i!(n - 1)!}` monomials of the form
    :math:`m_{i,k}(s) = \prod_{j=1}^n s_j^{p_{j,k}}, \ k = 1,\ldots,c_i` satisfying the
    condition :math:`\sum_{j=1}^n p_{j,k} = d_i` (see :func:`monomial_powers`).

    This function returns a basis vector
    :math:`\Phi : \mathbb{R}^n \rightarrow \mathbb{R}^{\sum_{i=0}^{M-m}{c_i}}` that
    stacks all of the aforementioned monomials:

    .. math::

        \Phi(s) = \Bigr[
            m_{0,1}(s), \ldots, m_{0,c_0}(s), m_{1,1}(s), \ldots, m_{1,c_1}(s), \ldots, m_{M-m,1}(s), \ldots, m_{M-m,c_{M-m}}(s)
        \Bigl]^\top

    Parameters
    ----------
    n : int
        Dimension of the input vector.
    mindegree : int
        Minimum degree of monomials (included).
    maxdegree : int
        Maximum degree of monomials (included).

    Returns
    -------
    :class:`casadi.Function`
        A casadi function of the form :math:`\Phi(s)`, where :math:`s \in \mathbb{R^n}`
        is the input and :math:`\Phi` the basis vector.

    Examples
    --------
    >>> import casadi as cs
    >>> from mpcrl.util.math import monomial_powers
    >>> Phi = monomials_basis_function(3, 1, 2)
    >>> s = cs.SX.sym("s", Phi.size1_in(0), 1)
    >>> Phi(s)
    SX([s_0, s_1, s_2, sq(s_0), (s_0*s_1), (s_0*s_2), sq(s_1), (s_1*s_2), sq(s_2)])
    """
    s = cs.SX.sym("s", n, 1)  # prod is faster with SX, and during runtime
    y = cs.vertcat(
        *(
            _prod(s**p)
            for k in range(mindegree, maxdegree + 1)
            for p in monomial_powers(n, k)
        )
    )
    return cs.Function("Phi", (s,), (y,), ("s",), ("Phi(s)",), {"cse": True})



from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np
import numpy.typing as npt

from .base_optimizer import BaseOptimizer, SymType


class GradientFreeOptimizer(BaseOptimizer[SymType], ABC):
    """Base class for gradient-free optimization algorithms, e.g., Bayesian
    Optimization.

    This optimizer adopts the ask-tell interface, i.e., it must implement the
    :meth:`GradientFreeOptimizer.ask` and :meth:`GradientFreeOptimizer.tell` methods.
    The former allows the agent to ask for a new set of parameters to evaluate, while
    the latter allows the agent to tell the optimizer the values of the objective
    function(s) for the set of parameters it asked for.
    """

    prefers_dict: bool
    """A flag that specifies whether the optimizer prefers to receive the learnable
    parameters as a dictionary of names and values or as a single concatenated array."""

    @abstractmethod
    def ask(
        self,
    ) -> tuple[Union[dict[str, npt.ArrayLike], npt.ArrayLike], Optional[str],]:
        """Asks the learning agent for a new set of parameters to evaluate.

        Returns
        -------
        dict of (str, 1d arrays) or a single 1d array
            A dictionary of learnable parameter names and their corresponding values.
            Or a single array that results from the concatenation of the parameter
            values.
        str, optional (default=None)
            A string that specifies the status of the optimizer. This is useful to
            communicate to the learning agent whether the optimization algorithm has
            encountered, e.g., some error or failure.
        """

    @abstractmethod
    def tell(
        self,
        values: Union[dict[str, npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        objective: Union[float, npt.NDArray[np.floating]],
    ) -> None:
        """Tells the learning agent the values of the objective function for the set of
        parameters it asked for.

        Parameters
        ----------
        values : dict of (str, 1d arrays) or a single 1d array
            A dictionary of learnable parameter names and their corresponding values for
            which the objective function(s) was (were) evaluated. Or a single array that
            results from the concatenation of the parameter values. This depends on the
            optimizer's :attr:`prefers_dict` class attribute.
        objective : float or array
            Value(s) of the objective function(s) for the set of parameters. Can be
            single-objective or multi-objective.
        """



import unittest

import casadi as cs
import numpy as np
from parameterized import parameterized, parameterized_class
from scipy.special import comb

from mpcrl.util import control, iters, math, named, seeding


class DummyAgent(named.Named):
    ...


class TestNamedAgent(unittest.TestCase):
    def test_init__with_given_name__saves_name_correctly(self):
        dummy = DummyAgent(name="ciao")
        self.assertEqual(dummy.name, "ciao")

    def test_init__without_name__creates_name_with_class(self):
        dummy0 = DummyAgent()
        dummy1 = DummyAgent()
        self.assertEqual(dummy0.name, "DummyAgent0")
        self.assertEqual(dummy1.name, "DummyAgent1")

    def test_str_and_repr(self):
        name = "ciao"
        dummy = DummyAgent(name=name)
        S = dummy.__str__()
        self.assertIn(name, S)
        S = dummy.__repr__()
        self.assertIn(name, S)
        self.assertIn(DummyAgent.__name__, S)


class TestMath(unittest.TestCase):
    def test_cholesky_added_multiple_identities__performs_correct_factorization(self):
        n = 5
        A = np.random.rand(n, n) + np.eye(n) * 5
        A = 0.5 * (A + A.T)
        L = math.cholesky_added_multiple_identity(A)
        np.testing.assert_allclose(A, L @ L.T)

    def test_cholesky_added_multiple_identities__raises__max_iterations_reached(self):
        n = 5
        A = np.random.rand(n, n) - np.eye(n) * 5
        A = 0.5 * (A + A.T)
        with self.assertRaisesRegex(ValueError, "Maximum iterations reached."):
            math.cholesky_added_multiple_identity(A, maxiter=1)

    @parameterized.expand(
        [
            (
                3,
                5,
                {
                    (0, 0, 5),
                    (0, 1, 4),
                    (0, 2, 3),
                    (0, 3, 2),
                    (0, 4, 1),
                    (0, 5, 0),
                    (1, 0, 4),
                    (1, 1, 3),
                    (1, 2, 2),
                    (1, 3, 1),
                    (1, 4, 0),
                    (2, 0, 3),
                    (2, 1, 2),
                    (2, 2, 1),
                    (2, 3, 0),
                    (3, 0, 2),
                    (3, 1, 1),
                    (3, 2, 0),
                    (4, 0, 1),
                    (4, 1, 0),
                    (5, 0, 0),
                },
            ),
            (1, 4, {(4,)}),
            (10, 1, {tuple(row) for row in np.eye(10)}),
            (4, 3, None),
        ]
    )
    def test_monomial_powers__computes_correct_powers(self, n, k, out):
        p = math.monomial_powers(n, k)
        np.testing.assert_array_equal(p.sum(axis=1), k)
        expected_n_combinations = comb(k + n - 1, n - 1, exact=True)
        self.assertEqual(p.shape, (expected_n_combinations, n))
        if out is None:
            return
        set_of_combinations = {tuple(row) for row in p}
        self.assertSetEqual(set_of_combinations, out)

    @parameterized.expand([(3, 1, 2), (5, 0, 5), (4, 5, 4)])
    def test_monomials_basis_function(self, n, mindeg, maxdeg):
        Phi = math.monomials_basis_function(n, mindeg, maxdeg)
        self.assertEqual(Phi.n_in(), 1)
        self.assertEqual(Phi.n_out(), 1)

        self.assertEqual(n, Phi.size1_in(0))
        expected_out = sum(
            comb(deg + n - 1, n - 1, exact=True) for deg in range(mindeg, maxdeg + 1)
        )
        self.assertEqual(expected_out, Phi.size1_out(0))


class TestControl(unittest.TestCase):
    def test_lqr__returns_correctly(self):
        K_exp = np.array([[0.307774887341029, 0.441249017604930]])
        P_exp = np.array(
            [
                [0.531874210236110, 0.035928068775907],
                [0.035928068775907, 0.422885782452801],
            ]
        )
        A = np.array([[-0.9, 0.25], [0, -1.1]])
        B = np.array([[0.23], [0.45]])
        Q = np.eye(2)
        R = np.atleast_2d(0.45)
        K, P = control.lqr(A, B, Q, R)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_lqr__with_M__returns_correctly(self):
        K_exp = np.array([[0.462006509187942, 0.766379710152564]])
        P_exp = np.array(
            [
                [0.502193051922548, -0.016892161794694],
                [-0.016892161794694, 0.330569037292075],
            ]
        )
        A = np.array([[-0.9, 0.25], [0, -1.1]])
        B = np.array([[0.23], [0.45]])
        Q = np.eye(2)
        R = np.atleast_2d(0.45)
        M = np.array([[0.1], [0.2]])
        K, P = control.lqr(A, B, Q, R, M)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_dlqr__returns_correctly(self):
        K_exp = np.array([[1.075936290787970, 1.824593914133278]])
        P_exp = np.array(
            [
                [6.783278637425703, 2.903698804441288],
                [2.903698804441288, 5.026668672843932],
            ]
        )
        A = np.array([[1, 0.25], [0, 1]])
        B = np.array([[0.03], [0.25]])
        Q = np.eye(2)
        R = 0.5
        K, P = control.dlqr(A, B, Q, R)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_dlqr__with_M__returns_correctly(self):
        K_exp = np.array([[1.132928272226698, 1.820247185260309]])
        P_exp = np.array(
            [
                [6.426698776552665, 2.359469798498852],
                [2.359469798498852, 3.806839220681791],
            ]
        )
        A = np.array([[1, 0.25], [0, 1]])
        B = np.array([[0.03], [0.25]])
        Q = np.eye(2)
        R = np.atleast_2d(0.5)
        M = np.array([[0.1], [0.2]])
        K, P = control.dlqr(A, B, Q, R, M)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_rk4__returns_correcly(self):
        def f(x):
            return -3 * x

        dt = 0.01
        x = cs.SX.sym("x")
        fd = cs.Function("fd", [x], [control.rk4(f, x, dt)])

        Y = [1]
        for _ in range(100):
            Y.append(fd(Y[-1]))
        Y = cs.hcat(Y).full().squeeze()
        Y_exact = np.exp(-3 * (np.arange(Y.size) * dt))
        np.testing.assert_allclose(Y, Y_exact)

    def test_cbf_degree_1(self):
        M = cs.SX.sym("M")
        v0 = cs.SX.sym("v0")

        p = cs.SX.sym("p")
        v = cs.SX.sym("v")
        z = cs.SX.sym("z")
        x = cs.vertcat(p, v, z)
        u = cs.SX.sym("u")
        friction_coeffs = cs.SX.sym("friction_coeffs", 3, 1)
        friction = cs.dot(friction_coeffs, cs.vertcat(1, v, v**2))
        x_dot = cs.vertcat(v, (u - friction) / M, v0 - v)
        dynamics = cs.Function("dynamics", [x, u], [x_dot], {"allow_free": True})

        Th = cs.SX.sym("Th")
        h = lambda x_: x_[2] - Th * x_[1]  # >= 0

        gamma = cs.SX.sym("gamma")
        alphas = [lambda w: gamma * w]
        cbf, _ = control.cbf(h, x, u, dynamics, alphas)

        actual_cbf = cbf(x, u)
        expected_cbf = Th / M * (friction - u) + (v0 - v) + gamma * (z - Th * v)
        self.assertTrue(all(cbf.shape == (1, 1) for cbf in [actual_cbf, expected_cbf]))

        variables = cs.symvar(actual_cbf)
        shapes = [v.shape for v in variables]
        diff = cs.Function("d", variables, (actual_cbf, expected_cbf))
        for _ in range(20):
            values = [np.random.randn(*shape) for shape in shapes]
            np.testing.assert_allclose(*diff(*values), atol=1e-12, rtol=1e-12)

    def test_cbf_degree_2(self):
        v = cs.SX.sym("v")
        z = cs.SX.sym("z")
        u = cs.SX.sym("u")
        v0 = cs.SX.sym("v0")
        x = cs.vertcat(v, z)
        f = cs.vertcat(0, v0 - v)
        g = cs.vertcat(1, 0)
        x_dot = f + g * u
        dynamics = cs.Function("dynamics", [x, u], [x_dot], {"allow_free": True})
        delta = cs.SX.sym("delta")
        h = lambda x_: x_[1] - delta  # >= 0

        # degree 1
        alphas = [lambda y: y]
        cbf1, _ = control.cbf(h, x, u, dynamics, alphas)
        actual_cbf1 = cbf1(x, u)
        expected_cbf1 = v0 - v + cs.SX.zeros(1, 1) * u + z - delta

        # degree 2
        alphas = [lambda y: y**2] * 2
        cbf2, _ = control.cbf(h, x, u, dynamics, alphas)
        actual_cbf2 = cbf2(x, u)
        h_ = h(x)
        Lfh_ = control.lie_derivative(h_, x, f)  # v0 - v
        Lf2h_ = control.lie_derivative(h_, x, f, 2)  # 0
        LgLfh_ = control.lie_derivative(Lfh_, x, g)  # -1
        expected_cbf2 = Lf2h_ + LgLfh_ * u + 2 * h_ * Lfh_ + (Lfh_ + h_**2) ** 2
        self.assertTrue(
            all(
                cbf.shape == (1, 1)
                for cbf in [actual_cbf1, expected_cbf1, actual_cbf2, expected_cbf2]
            )
        )

        variables = cs.symvar(actual_cbf2)
        shapes = [v.shape for v in variables]
        diff = cs.Function(
            "d", variables, (actual_cbf1, expected_cbf1, actual_cbf2, expected_cbf2)
        )
        for _ in range(200):
            values = [np.random.randn(*shape) for shape in shapes]
            out1, out2, out3, out4 = diff(*values)
            np.testing.assert_allclose(out1, out2, atol=1e-12, rtol=1e-12)
            np.testing.assert_allclose(out3, out4, atol=1e-12, rtol=1e-12)

    def test_dcbf_degree_1(self):
        A = cs.SX.sym("A", 2, 2)
        B = cs.SX.sym("B", 2, 1)
        x = cs.SX.sym("x", A.shape[0], 1)
        u = cs.SX.sym("u", B.shape[1], 1)
        dynamics = lambda x, u: A @ x + B @ u
        M = cs.SX.sym("M")
        c = cs.SX.sym("c")
        gamma = cs.SX.sym("gamma")
        alphas = [lambda z: gamma * z]
        h = lambda x: M - c * x[0]  # >= 0
        cbf, _ = control.dcbf(h, x, u, dynamics, alphas)

        actual_cbf = cbf(x, u)
        expected_cbf = -c * (A[0, :] @ x + B[0] * u - x[0]) + gamma * (M - c * x[0])
        self.assertTrue(all(cbf.shape == (1, 1) for cbf in [actual_cbf, expected_cbf]))

        variables = cs.symvar(actual_cbf)
        shapes = [v.shape for v in variables]
        diff = cs.Function("d", variables, (actual_cbf, expected_cbf))
        for _ in range(20):
            values = [np.random.randn(*shape) for shape in shapes]
            np.testing.assert_allclose(*diff(*values), atol=1e-12, rtol=1e-12)

    def test_dcbf_degree_2(self):
        A = cs.SX.sym("A", 2, 2)
        B = cs.SX.zeros(2, 1)
        B[1] = cs.SX.sym("b")
        x = cs.SX.sym("x", A.shape[0], 1)
        u = cs.SX.sym("u", B.shape[1], 1)
        dynamics = lambda x, u: A @ x + B @ u
        M = cs.SX.sym("M")
        h = lambda x: M - x[0]  # >= 0
        gamma = cs.SX.sym("gamma", 2, 1)

        # degree 1
        alphas = [lambda z: gamma[0] * z]
        cbf1, _ = control.dcbf(h, x, u, dynamics, alphas)
        actual_cbf1 = cbf1(x, u)
        expected_cbf1 = -A[0, :] @ x + x[0] + gamma[0] * (M - x[0])

        # degree 2
        alphas = [lambda z: gamma[0] * z, lambda z: gamma[1] * z]
        cbf2, _ = control.dcbf(h, x, u, dynamics, alphas)
        actual_cbf2 = cbf2(x, u)
        expected_cbf2 = (
            (A[0, :] @ x - x[0]) * (1 - A[0, 0] - cs.sum1(gamma))
            + (M - x[0]) * gamma[0] * gamma[1]
            - A[0, 1] * (A[1, :] @ x - x[1])
            - B[1] * A[0, 1] * u
        )
        self.assertTrue(
            all(
                cbf.shape == (1, 1)
                for cbf in [actual_cbf1, expected_cbf1, actual_cbf2, expected_cbf2]
            )
        )

        variables = cs.symvar(actual_cbf2)
        shapes = [v.shape for v in variables]
        diff = cs.Function(
            "d", variables, (actual_cbf1, expected_cbf1, actual_cbf2, expected_cbf2)
        )
        for _ in range(200):
            values = [np.random.randn(*shape) for shape in shapes]
            out1, out2, out3, out4 = diff(*values)
            np.testing.assert_allclose(out1, out2, atol=1e-12, rtol=1e-12)
            np.testing.assert_allclose(out3, out4, atol=1e-12, rtol=1e-12)


@parameterized_class("starts_with", [(False,), (True,)])
class TestIters(unittest.TestCase):
    @parameterized.expand([(5,), (1,), (22,)])
    def test_bool_cycle(self, frequency: int):
        T = frequency * 10
        cycle = iters.bool_cycle(frequency, self.starts_with)
        cycle = [next(cycle) for _ in range(T)]
        self.assertEqual(cycle[0], self.starts_with or frequency == 1)
        self.assertEqual(T // frequency, sum(cycle))


class TestSeeding(unittest.TestCase):
    def test_mk_seed(self):
        rng = np.random.default_rng()
        self.assertTrue(0 <= seeding.mk_seed(rng) < 2**32)


if __name__ == "__main__":
    unittest.main()



"""A submodule with utility functions and typing for seeding random number
generators. In particular, these have been conceived with the goal of being as
compatible with the :mod:`gymnasium` framework as possible."""

import sys
from collections.abc import Sequence as _Sequence
from typing import Union

import numpy as np

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

RngType: TypeAlias = Union[
    None,
    int,
    _Sequence[int],
    np.random.SeedSequence,
    np.random.BitGenerator,
    np.random.Generator,
]


MAX_SEED = np.iinfo(np.uint32).max + 1


def mk_seed(rng: np.random.Generator) -> int:
    """Generates a random seed compatible with :func:`gymnasium.Env.reset`.

    Parameters
    ----------
    rng : :class:`numpy.random.Generator`
        RNG generator.

    Returns
    -------
    seed
        A random integer in the range [0, 2**32).
    """
    return int(rng.integers(MAX_SEED))



from typing import Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.schedulers import Scheduler
from .gradient_based_optimizer import GradientBasedOptimizer, LrType, SymType


class GradientDescent(GradientBasedOptimizer[SymType, LrType]):
    r"""First-order Gradient descent optimizer, based on
    :cite:`sutskever_importance_2013` and :class:`torch.optim.SGD`.

    In its basic formulation, this optimizer updates the parameters as

    .. math:: \theta \gets \theta - \alpha g,

    where :math:`\theta` are the learnable parameters, :math:`\alpha` is the learning
    rate (could be extended to the case this is a vector of rates), and :math:`g` is the
    gradient of the loss function w.r.t. the parameters. If momentum or weight decay are
    used, the gradient :math:`g` is modified before using it, but the update rule
    remains the same. However, when considering a constrained parameter space, we need
    to solve a Quadratic Programming (QP) problem to ensure the parameters stay within
    their bounds. For gradient descent, the QP problem is

    .. math::

        \begin{aligned}
            \min_{\Delta\theta} & \quad \frac{1}{2} \lVert \Delta\theta \rVert_2^2 + \alpha g^\top \Delta\theta \\
            \text{s.t.} & \quad \theta_{\text{lower}} \leq \theta + \Delta\theta \leq \theta_{\text{upper}}
        \end{aligned}

    followed by the update :math:`\theta \gets \theta + \Delta\theta`.

    Parameters
    ----------
    learning_rate : float or array or :class:`mpcrl.core.schedulers.Scheduler`
        The learning rate of the optimizer. It can be:

        - a float, in case the learning rate must stay constant and is the same for all
          learnable parameters

        - an array, in case the learning rate must stay constant but is different for
          each parameter (should have the same size as the number of learnable
          parameters)

        - a :class:`mpcrl.core.schedulers.Scheduler`, in case the learning rate can vary
          during the learning process (usually, it is set to decay). See the ``hook``
          argument for more details on when this scheduler is stepped.
    weight_decay : float, optional
        A positive float that specifies the decay of the learnable parameters in the
        form of an L2 regularization term. By default, it is set to ``0.0``, so no
        decay/regularization takes place.
    momentum : float, optional
        A positive float that specifies the momentum factor. By default, it is set
        to ``0.0``, so no momentum is used.
    dampening : float, optional
        A positive float that specifies the dampening factor for the momentum. By
        default, it is set to ``0.0``, so no dampening is used.
    nesterov : bool, optional
        A boolean that specifies whether to use Nesterov momentum. By default, it is
        set to ``False``.
    hook : {"on_update", "on_episode_end", "on_timestep_end"}, optional
        Specifies when to step the optimizer's learning rate's scheduler to decay
        its value. This allows to vary the rate over the learning iterations. The
        options are:

        - ``"on_update"`` steps the learning rate after each agent's update

        - ``"on_episode_end"`` steps the learning rate after each episode's end

        - ``"on_timestep_end"`` steps the learning rate after each env's timestep.

        By default, ``"on_update"`` is selected.
    max_percentage_update : float, optional
        A positive float that specifies the maximum percentage change the learnable
        parameters can experience in each update. For example,
        ``max_percentage_update=0.5`` means that the parameters can be updated by up
        to 50% of their current value. By default, it is set to ``+inf``.
        If specified, the update becomes constrained and has to be solved as a QP, which
        is inevitably slower than its unconstrained counterpart (a linear system).
    bound_consistency : bool, optional
        A boolean that, if ``True``, forces the learnable parameters to lie in their
        bounds when updated. This is done via :func:`numpy.clip`. Only beneficial if
        numerical issues arise during updates, e.g., due to the QP solver not being able
        to guarantee bounds.
    """

    _order = 1
    _hessian_sparsity = "diag"
    # In GD, the hessian is at most diagonal, i.e., in case we have constraints.

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        weight_decay: float = 0.0,
        momentum: float = 0.0,
        dampening: float = 0.0,
        nesterov: bool = False,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
        bound_consistency: bool = False,
    ) -> None:
        super().__init__(learning_rate, hook, max_percentage_update, bound_consistency)
        self.weight_decay = weight_decay
        self.momentum = momentum
        self.dampening = dampening
        self.nesterov = nesterov
        self._momentum_buffer = None

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value

        # compute candidate update
        dtheta, self._momentum_buffer = _gd(
            theta,
            gradient,
            self._momentum_buffer,
            self.weight_decay,
            self.momentum,
            self.lr_scheduler.value,
            self.dampening,
            self.nesterov,
        )

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            return theta + dtheta, None
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=cs.DM.eye(theta.shape[0]), g=-dtheta, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _gd(
    theta: np.ndarray,
    g: np.ndarray,
    momentum_buffer: Optional[np.ndarray],
    weight_decay: float,
    momentum: float,
    lr: LrType,
    dampening: float,
    nesterov: bool,
) -> tuple[np.ndarray, Optional[np.ndarray]]:
    """Computes the update's change according to the gradient descent algorithm."""
    if weight_decay > 0.0:
        g += weight_decay * theta
    if momentum > 0.0:
        if momentum_buffer is None:
            momentum_buffer = g.copy()
        else:
            momentum_buffer = momentum * momentum_buffer + (1 - dampening) * g
        if nesterov:
            g += momentum * momentum_buffer
        else:
            g = momentum_buffer
    dtheta = -lr * g
    return dtheta, momentum_buffer



import logging
import os
import unittest
from functools import lru_cache
from itertools import combinations, product
from random import random
from typing import Any
from unittest.mock import Mock, call

import casadi as cs
import gymnasium as gym
import numpy as np
from csnlp import Nlp, scaling
from csnlp.multistart import StackedMultistartNlp
from csnlp.wrappers import Mpc, NlpScaling
from parameterized import parameterized

from mpcrl import (
    Agent,
    LearnableParameter,
    LearnableParametersDict,
    LearningAgent,
    wrappers_agents,
    wrappers_envs,
)
from mpcrl.wrappers.envs.monitor_infos import _compact_dicts


@lru_cache
def get_dynamics(g: float, alpha: float, dt: float) -> cs.Function:
    x, u, d = cs.SX.sym("x", 3), cs.SX.sym("u", 2), cs.SX.sym("d", 3)
    x_next = x + cs.vertcat(x[1], u[0] / x[2] - g, -alpha * u[0]) * dt + d * 0
    return cs.Function("F", [x, u, d], [x_next], ["x", "u", "d"], ["x+"])


def get_mpc(horizon: int, multistart: bool):
    N = horizon
    T = 100
    K = 3
    dt = T / N
    yT = 100000
    g = 9.81
    alpha = 1 / (300 * g)
    y_nom = 1e5
    v_nom = 2e3
    m_nom = 3e5
    u_nom = 1e8
    scaler = scaling.Scaler()
    scaler.register("y", scale=y_nom)
    scaler.register("y_0", scale=y_nom)
    scaler.register("v", scale=v_nom)
    scaler.register("v_0", scale=v_nom)
    scaler.register("m", scale=m_nom)
    scaler.register("m_0", scale=m_nom)
    scaler.register("u1", scale=u_nom)
    scaler.register("u2", scale=u_nom)
    nlp = (
        StackedMultistartNlp[cs.MX](sym_type="MX", starts=K)
        if multistart
        else Nlp[cs.MX](sym_type="MX")
    )
    nlp = NlpScaling[cs.MX](nlp, scaler=scaler)
    mpc = Mpc[cs.MX](nlp, prediction_horizon=N)
    y, _ = mpc.state("y")
    _, _ = mpc.state("v")
    m, _ = mpc.state("m", lb=0)
    mpc.action("u1", lb=0, ub=5e7)
    u2, _ = mpc.action("u2", lb=0, ub=5e7)
    mpc.disturbance("d", 3)
    mpc.set_dynamics(get_dynamics(g, alpha, dt))
    mpc.constraint("yT", y[-1], "==", yT)
    mpc.minimize(m[0] - m[-1] + cs.sum2(u2))
    mpc.init_solver()
    return mpc


class DummyAgent(Agent):
    ...


class DummyLearningAgent(LearningAgent):
    def update(self, *args, **kwargs):
        return

    def train_one_episode(self, *args, **kwargs):
        return


def mk_agent() -> DummyLearningAgent:
    return DummyLearningAgent(
        mpc=get_mpc(3, False),
        learnable_parameters=LearnableParametersDict(),
        update_strategy=1,
    )


class SimpleEnv(gym.Env[object, object]):
    T_MAX = 10

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.INTERNAL_RESET_INFOS = []
        self.INTERNAL_FINALIZED_RESET_INFOS = {"reset_info": []}
        self.INTERNAL_STEP_INFOS = []
        self.INTERNAL_FINALIZED_STEP_INFOS = {"step_info": []}
        self.INTERNAL_OBSERVATIONS = []
        self.INTERNAL_ACTIONS = []
        self.INTERNAL_REWARDS = []

    def reset(self, *args: Any, **kwargs: Any) -> tuple[object, dict[str, Any]]:
        super().reset(*args, **kwargs)
        self.t = 0
        obs = object()
        info = {"reset_info": object()}
        self.INTERNAL_RESET_INFOS.append(info)
        self.INTERNAL_FINALIZED_RESET_INFOS["reset_info"].append(info["reset_info"])
        self.INTERNAL_STEP_INFOS.append([])
        self.INTERNAL_FINALIZED_STEP_INFOS["step_info"].append([])
        self.INTERNAL_OBSERVATIONS.append([obs])
        self.INTERNAL_ACTIONS.append([])
        self.INTERNAL_REWARDS.append([])
        return obs, info

    def step(self, action: object) -> tuple[object, float, bool, bool, dict[str, Any]]:
        self.t += 1
        obs = object()
        reward = np.random.rand()
        terminated = self.t >= self.T_MAX
        info = {"step_info": object()}
        self.INTERNAL_STEP_INFOS[-1].append(info)
        self.INTERNAL_FINALIZED_STEP_INFOS["step_info"][-1].append(info["step_info"])
        self.INTERNAL_OBSERVATIONS[-1].append(obs)
        self.INTERNAL_ACTIONS[-1].append(action)
        self.INTERNAL_REWARDS[-1].append(reward)
        return obs, reward, terminated, False, info


class TestWrapperAndLearningWrapper(unittest.TestCase):
    def test_attr__raises__when_accessing_private_attrs(self):
        wrapped = wrappers_agents.LearningWrapper(mk_agent())
        with self.assertRaisesRegex(
            AttributeError, "Accessing private attribute '_x' is prohibited."
        ):
            wrapped._x

    def test_unwrapped__unwraps_nlp_correctly(self):
        agent = mk_agent()
        wrapped = wrappers_agents.LearningWrapper(agent)
        self.assertIs(agent, wrapped.unwrapped)

    def test_str_and_repr(self):
        agent = mk_agent()
        wrapped = wrappers_agents.LearningWrapper(agent)
        S = wrapped.__str__()
        self.assertIn(wrappers_agents.LearningWrapper.__name__, S)
        self.assertIn(agent.__str__(), S)
        S = wrapped.__repr__()
        self.assertIn(wrappers_agents.LearningWrapper.__name__, S)
        self.assertIn(agent.__repr__(), S)

    def test_is_wrapped(self):
        agent = mk_agent()
        self.assertFalse(agent.is_wrapped())

        wrapped = wrappers_agents.LearningWrapper(agent)
        self.assertTrue(wrapped.is_wrapped(wrappers_agents.LearningWrapper))
        self.assertFalse(wrapped.is_wrapped(cs.SX))

    def test_detach_wrapper(self):
        cp = lambda d: {k: {k_: id(v_) for k_, v_ in v.items()} for k, v in d.items()}
        agent_0 = mk_agent()
        agent_0_hooks = cp(agent_0._hooks)
        wrapped_1 = wrappers_agents.RecordUpdates(agent_0)
        wrapped_1_hooks = cp(agent_0._hooks)
        wrapped_2 = wrappers_agents.Log(
            wrapped_1, level=logging.DEBUG, log_frequencies={"on_timestep_end": 1000}
        )
        wrapped_2_hooks = cp(agent_0._hooks)
        for d1, d2 in combinations(
            (agent_0_hooks, wrapped_1_hooks, wrapped_2_hooks), 2
        ):
            with self.assertRaises(AssertionError):
                self.assertDictEqual(d1, d2)

        detached_1 = wrapped_2.detach_wrapper()
        self.assertIs(detached_1, wrapped_1)
        self.assertDictEqual(cp(agent_0._hooks), wrapped_1_hooks)

        detached_0 = detached_1.detach_wrapper()
        self.assertIs(detached_0, agent_0)
        self.assertDictEqual(cp(agent_0._hooks), agent_0_hooks)

        detached_recursive = wrapped_2.detach_wrapper(recursive=True)
        self.assertIs(detached_recursive, agent_0)
        self.assertDictEqual(cp(agent_0._hooks), agent_0_hooks)


class TestLog(unittest.TestCase):
    def setUp(self):
        self.agent = mk_agent()
        self.env = SimpleEnv()
        self.name = "test_logger"
        self.log = None

    def tearDown(self) -> None:
        # remove logging files with name "DummyLearningAgent*.txt"
        for handler in self.log.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.close()
                if os.path.isfile(handler.baseFilename):
                    os.remove(handler.baseFilename)

    def test_init__to_file__creates_log_file(self):
        self.log = wrappers_agents.Log(self.agent, to_file=True)
        found = False
        for handler in self.log.logger.handlers:
            if isinstance(
                handler, logging.FileHandler
            ) and handler.baseFilename.endswith(f"{self.agent.name}.txt"):
                found = True
                break
        self.assertTrue(found)

    def test_init__establishes_correct_hooks(self):
        agent = DummyAgent(get_mpc(3, False))
        log_frequencies = {
            "on_episode_start": 2,
            "on_episode_end": 3,
            "on_env_step": 4,
            "on_timestep_end": 5,
            "on_update": 6,
        }
        exclude_mandatory = ["on_mpc_failure", "on_validation_start"]
        self.log = wrappers_agents.Log(
            agent, log_frequencies=log_frequencies, exclude_mandatory=exclude_mandatory
        )

        expected_hooks = {
            "on_episode_start",
            "on_episode_end",
            "on_env_step",
            "on_timestep_end",
            "on_validation_end",
        }
        actual_hooks = set(self.log.unwrapped._hooks.keys())
        self.assertEqual(len(set.difference(expected_hooks, actual_hooks)), 0)

    def test_on_mpc_failure(self):
        episode = 1
        timestep = 10
        status = "failed"
        raises = False
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.WARNING):
            self.log.on_mpc_failure(episode, timestep, status, raises)

    def test_on_validation_start(self):
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.DEBUG):
            self.log.on_validation_start(self.env)

    def test_on_validation_end(self):
        returns = np.array([1.0, 2.0, 3.0])
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.INFO):
            self.log.on_validation_end(self.env, returns)

    def test_on_episode_start(self):
        episode = 1
        state = np.array([1, 2, 3])
        self.log = wrappers_agents.Log(
            self.agent, log_frequencies={"on_episode_start": 1}
        )
        with self.assertLogs(self.log.logger, logging.DEBUG):
            self.log.on_episode_start(self.env, episode, state)

    def test_on_episode_end(self):
        episode = 1
        rewards = 10.0
        self.log = wrappers_agents.Log(
            self.agent, log_frequencies={"on_episode_end": 1}
        )
        with self.assertLogs(self.log.logger, logging.INFO):
            self.log.on_episode_end(self.env, episode, rewards)

    def test_on_env_step(self):
        episode = 1
        timestep = 10
        self.log = wrappers_agents.Log(self.agent, log_frequencies={"on_env_step": 1})
        with self.assertLogs(self.log.logger, logging.DEBUG):
            self.log.on_env_step(self.env, episode, timestep)

    def test_on_timestep_end(self):
        episode = 1
        timestep = 10
        self.log = wrappers_agents.Log(
            self.agent, log_frequencies={"on_timestep_end": 1}
        )
        with self.assertLogs(self.log.logger, logging.DEBUG):
            self.log.on_timestep_end(self.env, episode, timestep)

    def test_on_update_failure(self):
        episode = 1
        timestep = 10
        errormsg = "update failed"
        raises = False
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.WARNING):
            self.log.on_update_failure(episode, timestep, errormsg, raises)

    def test_on_training_start(self):
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.DEBUG):
            self.log.on_training_start(self.env)

    def test_on_training_end(self):
        returns = np.array([1.0, 2.0, 3.0])
        self.log = wrappers_agents.Log(self.agent)
        with self.assertLogs(self.log.logger, logging.INFO):
            self.log.on_training_end(self.env, returns)

    def test_on_update(self):
        self.log = wrappers_agents.Log(self.agent, log_frequencies={"on_update": 1})
        with self.assertLogs(self.log.logger, logging.INFO):
            self.log.on_update()


class TestRecordUpdates(unittest.TestCase):
    def test__update__records_learnable_parameters_correctly(self):
        n_params, n_updates = 3, 10
        pars = np.random.rand(n_updates + 1, n_params)
        parsdict = LearnableParametersDict(
            [LearnableParameter(f"p{i}", 1, pars[0, i]) for i in range(n_params)]
        )
        agent = mk_agent()
        agent._learnable_pars = parsdict
        on_update_original = agent.on_update
        pars_iter = iter(pars[1:])

        def side_effect():
            agent.learnable_parameters.update_values(next(pars_iter))
            on_update_original()

        agent.on_update = Mock(return_value=None, side_effect=side_effect)
        wrapped = wrappers_agents.RecordUpdates(agent)

        for _ in range(n_updates):
            wrapped.on_update()

        agent.on_update.assert_has_calls([call()] * n_updates)
        self.assertListEqual(
            [f"p{i}" for i in range(n_params)], list(wrapped.updates_history.keys())
        )
        pars_actual = np.squeeze(list(wrapped.updates_history.values())).T
        np.testing.assert_equal(pars_actual, pars)


class TestEvaluate(unittest.TestCase):
    @parameterized.expand([(False,), (True,)])
    def test_evaluate__does_not_evaluate_if_not_training(self, is_training):
        frequency = 10
        repeats = 2
        agent = mk_agent()
        agent._is_training = is_training
        agent.evaluate = Mock()
        env = SimpleEnv()
        _ = wrappers_agents.Evaluate(agent, env, "on_episode_end", frequency=frequency)

        n_calls = frequency * repeats
        _ = [agent.on_episode_end(env, i, random()) for i in range(n_calls)]

        if is_training:
            self.assertGreater(agent.evaluate.call_count, 0)
        else:
            self.assertEqual(agent.evaluate.call_count, 0)

    @parameterized.expand(product((False, True), (False, True)))
    def test_evaluate__evaluates_with_correct_frequency(
        self, eval_immediately: bool, fix_seed: bool
    ):
        frequency = 10
        repeats = 2
        returns = [object() for _ in range(repeats + eval_immediately)]
        returns_iter = iter(returns)
        agent = mk_agent()
        agent._is_training = True  # otherwise, Evaluate will not be invoked
        agent.evaluate = Mock(side_effect=lambda *_, **__: next(returns_iter))
        env = SimpleEnv()
        wrapped = wrappers_agents.Evaluate(
            agent,
            env,
            "on_episode_end",
            frequency=frequency,
            eval_immediately=eval_immediately,
            fix_seed=fix_seed,
        )

        n_calls = frequency * repeats
        n_calls += int(frequency / 2)  # adds some spurious calls
        _ = [agent.on_episode_end(env, i, random()) for i in range(n_calls)]

        self.assertEqual(agent.evaluate.call_count, repeats + eval_immediately)
        self.assertListEqual(wrapped.eval_returns, returns)
        if fix_seed:
            seeds = (call.args[3] for call in agent.evaluate.call_args_list)
            self.assertEqual(len(set(seeds)), 1)


class TestMonitorEpisodesAndInfos(unittest.TestCase):
    def test__compact_dicts(self):
        act = _compact_dicts(
            [
                {"a": 3, "b": 2},
                {"b": 3, "c": 2},
                {"a": 3, "c": 2},
                {"c": 3, "d": 2},
                {"a": 3, "d": 2},
            ],
            fill_value=np.nan,
        )
        exp = {
            "a": [3, np.nan, 3, np.nan, 3],
            "b": [2, 3, np.nan, np.nan, np.nan],
            "c": [np.nan, 2, 2, 3, np.nan],
            "d": [np.nan, np.nan, np.nan, 2, 2],
        }
        self.assertDictEqual(act, exp)

    def test_monitor_infos__records_infos_correctly(self):
        env = wrappers_envs.MonitorInfos(SimpleEnv())
        n_episodes = 3
        for _ in range(n_episodes):
            env.reset()
            terminated = truncated = False
            while not (terminated or truncated):
                _, _, terminated, truncated, _ = env.step(object())
        self.assertListEqual(
            list(env.reset_infos), env.get_wrapper_attr("INTERNAL_RESET_INFOS")
        )
        self.assertDictEqual(
            env.finalized_reset_infos(),
            env.get_wrapper_attr("INTERNAL_FINALIZED_RESET_INFOS"),
        )
        self.assertListEqual(
            list(env.step_infos), env.get_wrapper_attr("INTERNAL_STEP_INFOS")
        )
        self.assertDictEqual(
            env.finalized_step_infos(),
            env.get_wrapper_attr("INTERNAL_FINALIZED_STEP_INFOS"),
        )

    def test_monitor_episodes__records_episodes_correctly(self):
        env = wrappers_envs.MonitorEpisodes(SimpleEnv())
        n_episodes = 3
        for _ in range(n_episodes):
            env.reset()
            terminated = truncated = False
            while not (terminated or truncated):
                _, _, terminated, truncated, _ = env.step(object())
        np.testing.assert_array_equal(
            env.observations, env.get_wrapper_attr("INTERNAL_OBSERVATIONS")
        )
        np.testing.assert_array_equal(
            env.actions, env.get_wrapper_attr("INTERNAL_ACTIONS")
        )
        np.testing.assert_array_equal(
            env.rewards, env.get_wrapper_attr("INTERNAL_REWARDS")
        )
        self.assertListEqual(
            list(env.episode_lengths), [env.get_wrapper_attr("T_MAX")] * n_episodes
        )


if __name__ == "__main__":
    unittest.main()
